import collections
import contextlib
import operator
import ctypes
import queue
import multiprocessing
import typing
import os
import threading

import psutil

from ..easel import Alphabet, DigitalSequence, DigitalMSA, DigitalSequenceBlock, SequenceFile
from ..plan7 import TopHits, Builder, Pipeline, HMM, Profile, OptimizedProfile, HMMPressedFile, OptimizedProfileBlock
from ..utils import singledispatchmethod, peekable
from ._base import _BaseDispatcher, _BaseWorker, _BaseChore, _AnyProfile

_SEARCHQueryType = typing.Union[_AnyProfile]
_P = typing.TypeVar("_P", HMM, Profile, OptimizedProfile)

if typing.TYPE_CHECKING:
    from ._base import Unpack, PipelineOptions, BACKEND

# --- Worker -------------------------------------------------------------------

class _SEARCHWorker(
    _BaseWorker[
        _SEARCHQueryType,
        typing.Union[DigitalSequenceBlock, "SequenceFile[DigitalSequence]"],
        "TopHits[_SEARCHQueryType]",
    ],
):
    @singledispatchmethod
    def query(self, query) -> "TopHits[Any]":  # type: ignore
        raise TypeError(
            "Unsupported query type for `hmmsearch`: {}".format(type(query).__name__)
        )

    @query.register(HMM)
    @query.register(Profile)
    @query.register(OptimizedProfile)
    def _(self, query: _AnyProfile) -> "TopHits[_AnyProfile]":  # type: ignore
        assert self.pipeline is not None
        return self.pipeline.search_hmm(query, self.targets)


class _SEARCHThread(_SEARCHWorker, threading.Thread):
    pass


class _SEARCHProcess(_SEARCHWorker, multiprocessing.Process):
    pass


# --- Dispatcher ---------------------------------------------------------------

class _SEARCHDispatcher(
    _BaseDispatcher[
        _SEARCHQueryType,
        typing.Union[DigitalSequenceBlock, "SequenceFile[DigitalSequence]"],
        "TopHits[_SEARCHQueryType]",
    ]
):
    def _new_worker(
        self,
        query_queue: "queue.Queue[typing.Optional[_BaseChore[_SEARCHQueryType, TopHits[_SEARCHQueryType]]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        kill_switch: threading.Event,
    ) -> _SEARCHWorker:
        if isinstance(self.targets, SequenceFile):
            assert self.targets.name is not None
            targets = SequenceFile(
                self.targets.name,
                format=self.targets.format,
                digital=True,
                alphabet=self.options["alphabet"],
            )
        else:
            targets = self.targets  # type: ignore
        params = [
            targets,
            query_queue,
            query_count,
            kill_switch,
            self.callback,
            self.options,
            self.builder,
        ]
        if self.backend == "threading":
            return _SEARCHThread(*params)
        elif self.backend == "multiprocessing":
            return _SEARCHProcess(*params)
        else:
            raise ValueError(f"Invalid backend for `hmmsearch`: {self.backend!r}")


class _ReverseSEARCHDispatcher(
    _BaseDispatcher[
        "_SEARCHQueryType",
        DigitalSequenceBlock,
        "TopHits[_SEARCHQueryType]",
    ]
):
    """A ``hmmsearch`` dispatcher that parallelizes on the targets.
    """

    def __init__(
        self,
        queries: typing.Iterable["_SEARCHQueryType"],
        targets: DigitalSequenceBlock,
        cpus: int = 0,
        callback: typing.Optional[typing.Callable[["_SEARCHQueryType", int], None]] = None,
        builder: typing.Optional[Builder] = None,
        timeout: int = 1,
        backend: "BACKEND" = "threading",
        **options,  # type: Unpack[PipelineOptions]
    ) -> None:
        super().__init__(
            queries,
            targets,
            cpus,
            callback,
            builder,
            timeout,
            backend,
            **options
        )
        # only use as many CPUs as there are targets (if only a few), but
        # that may be a waste for less than N sequences per CPUs?
        self.cpus = max(1, min(cpus, len(targets)))
        # attempt to balance the chunks so that every thread gets about the
        # same number of *residues* (not the same number of *sequences*!)
        self.target_chunks = self._make_chunks(targets)

    def _make_chunks(self, targets: DigitalSequenceBlock) -> typing.List[DigitalSequenceBlock]:
        # compute chunksize from total sequence lengths
        # TODO: implement this as a Cython function with quick access to the
        #       sequence data?
        total_length = sum(len(seq) for seq in targets)
        chunksize = (total_length + self.cpus - 1) // self.cpus
        # balance sequence residues across chunks
        current_size = 0
        chunk_indices = [0]
        for i, seq in enumerate(targets):
            current_size += len(seq)
            if current_size > chunksize:
                chunk_indices.append(i)
                current_size = 0
        while len(chunk_indices) <= self.cpus:
            chunk_indices.append(len(targets))
        # NB: this does not copy data, as `DigitalSequenceBlock` are implemented
        #     as views of `DigitalSequence` objects, so slicing is cheap.
        return [targets[i:j] for i,j in zip(chunk_indices, chunk_indices[1:])]

    def _new_worker(
        self,
        query_queue,
        query_count,
        kill_switch: threading.Event,
        targets: DigitalSequenceBlock = None,
    ):
        if targets is None:
            targets = self.targets
        params = [
            targets,
            query_queue,
            query_count,
            kill_switch,
            None,
            self.options,
            self.builder,
        ]
        if self.backend == "threading":
            return _SEARCHThread(*params)
        elif self.backend == "multiprocessing":
            return _SEARCHProcess(*params)
        else:
            raise ValueError(f"Invalid backend for `hmmsearch`: {self.backend!r}")

    def _single_threaded(self) -> typing.Iterator["TopHits[_SEARCHQueryType]"]:
        for i, hits in enumerate(super()._single_threaded(), start=1):
            if self.callback is not None:
                self.callback(hits.query, i)
            yield hits

    def _multi_threaded(self) -> typing.Iterator["TopHits[_SEARCHQueryType]"]:
        with contextlib.ExitStack() as ctx:
            # single shared kill switch and query count
            if self.backend == "multiprocessing":
                manager = ctx.enter_context(multiprocessing.Manager())
                query_count = manager.Value(ctypes.c_ulong, 0)
                kill_switch = manager.Event()
            else:
                query_count = multiprocessing.Value(ctypes.c_ulong)  # type: ignore
                kill_switch = threading.Event()

            # create and launch one pipeline thread per CPU, each with its own
            # queue as they all need to get the same copy of each query
            workers = []
            queues = []
            for i in range(self.cpus):
                # create the queues to pass the query objects around, only
                # one item at a time since we synchronize query-by-query
                results: typing.Deque[_BaseChore[_Q, _R]] = collections.deque()
                if self.backend == "multiprocessing":
                    query_queue = ctx.enter_context(contextlib.closing(multiprocessing.Queue()))
                elif self.backend == "threading":
                    query_queue = queue.Queue()
                # create worker
                worker = self._new_worker(query_queue, query_count, kill_switch, targets=self.target_chunks[i])
                worker.start()
                workers.append(worker)
                queues.append(query_queue)

            # catch exceptions to kill threads in the background before exiting
            try:
                hits = None
                # process queries sequentially
                for i, query in enumerate(self.queries):
                    # yield hits obtained at the previous iteration
                    if hits is not None:
                        yield hits
                    query_count.value += 1
                    # create one chore per worker
                    chores = []
                    for worker, worker_queue in zip(workers, queues):
                        if isinstance(query, OptimizedProfile):
                            chore = self._new_chore(query.copy())
                        else:
                            chore = self._new_chore(query)
                        chores.append(chore)
                        worker_queue.put(chore)
                    # collect hits
                    partial_hits = []
                    for chore in chores:
                        partial_hits.append(chore.get())
                    # merge hits
                    hits = TopHits.merge(*partial_hits)
                    # call callback here after the hits have been merged
                    if self.callback is not None:
                        self.callback(chore.query, query_count.value)
                # now that we exhausted all queries, poison pill the
                # threads so they stop on their own gracefully
                for worker in workers:
                    worker.query_queue.put(None)
                    worker.join()
                    if self.backend == "multiprocessing":
                        worker.query_queue.close()
                # yield the final hits
                if hits is not None:
                    yield hits
            except BaseException as e:
                # make sure threads are killed to avoid being stuck,
                # e.g. after a KeyboardInterrupt, then re-raise
                try:
                    kill_switch.set()
                except queue.Full:
                    pass
                for worker in workers:
                    worker.join()
                    if self.backend == "multiprocessing":
                        worker.query_queue.close()
                        worker.query_queue.join_thread()
                raise e


# --- hmmsearch --------------------------------------------------------------

def hmmsearch(
    queries: typing.Union[_P, typing.Iterable[_P]],
    sequences: typing.Iterable[DigitalSequence],
    *,
    cpus: int = 0,
    callback: typing.Optional[typing.Callable[[_P, int], None]] = None,
    backend: "BACKEND" = "threading",
    parallel: "PARALLEL" = None,
    **options,  # type: Unpack[PipelineOptions]
) -> typing.Iterator["TopHits[_P]"]:
    """Search HMM profiles against a sequence database.

    In HMMER many-to-many comparisons, a *search* is the operation of
    querying with profile HMMs a database of sequences.

    The `hmmsearch` function offers two ways of managing the database that
    will be selected based on the type of the ``sequences`` argument. If
    ``sequences`` is an `SequenceFile` object, `hmmsearch` will reopen the
    file in each thread, and load targets *iteratively* to scan with the
    query. Otherwise, it will *pre-fetch* the target sequences into a
    `DigitalSequenceBlock` collection, and share them across threads
    without copy. The *pre-fetching* gives much higher performance at the
    cost of extra  startup time and much higher memory consumption. You may
    want to check how much memory is available (for instance with
    `psutil.virtual_memory`) before trying to load a whole sequence database,
    but it is really recommended to do so whenever possible.

    Arguments:
        queries (iterable of `HMM`, `Profile` or `OptimizedProfile`): The
            query HMMs or profiles to search for in the database. Note that
            passing a single object is supported.
        sequences (iterable of `~pyhmmer.easel.DigitalSequence`): A
            database of sequences to query. If you plan on using the
            same sequences several times, consider storing them into
            a `~pyhmmer.easel.DigitalSequenceBlock` directly. If a
            `~pyhmmer.easel.SequenceFile` is given, profiles will be loaded
            iteratively from disk rather than prefetched.
        cpus (`int`): The number of threads to run in parallel. Pass ``1``
            to run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        callback (callable): A callback that is called everytime a query is
            processed with two arguments: the query, and the total number
            of queries. This can be used to display progress in UI.
        backend (`str`): The parallel backend to use for workers to be
            executed. Supports ``threading`` to use thread-based parallelism,
            or ``multiprocessing`` to use process-based parallelism.
        parallel (`str`): The parallel strategy to use. Supports ``queries``
            to run queries in parallel, or ``targets`` to parallelize on
            targets while running one query at a time. If `None` given,
            use ``queries`` by default unless we can detect that there is
            a single or a small number of queries. Note that parallelization
            on ``targets`` does not work with `~pyhmmer.easel.SequenceFile`
            targets.

    Yields:
        `~pyhmmer.plan7.TopHits`: An object reporting *top hits* for each
        query, in the same order the queries were passed in the input.

    Raises:
        `~pyhmmer.errors.AlphabetMismatch`: When any of the query HMMs
            and the sequences do not share the same alphabet.
        `RuntimeError`: When attempting to use ``targets`` parallel
            strategy with targets from a `~pyhmmer.easel.SequenceFile`.

    Note:
        Any additional arguments passed to the `hmmsearch` function will be
        passed transparently to the `~pyhmmer.plan7.Pipeline` to be created.
        For instance, to run a ``hmmsearch`` using a bitscore cutoffs of
        5 instead of the default E-value cutoff, use::

            >>> hits = next(hmmsearch(thioesterase, proteins, T=5))
            >>> hits[0].score
            8.601...

        Since *version 0.11.0*, ``mypy`` should be able to detection which
        keywords can be passed to `hmmsearch` using a `TypedDict` annotation.

    .. versionadded:: 0.1.0

    .. versionadded:: 0.11.1
       ``parallel`` argument to select parallelization strategy.

    .. versionchanged:: 0.4.9
       Allow using `Profile` and `OptimizedProfile` queries.

    .. versionchanged:: 0.7.0
        Queries may now be an iterable of different types, or a single object.

    """
    cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or os.cpu_count() or 1
    alphabet = options.get("alphabet")

    if not isinstance(queries, collections.abc.Iterable):
        queries = (queries,)

    if isinstance(sequences, SequenceFile):
        if sequences.name is None:
            raise ValueError("expected named `SequenceFile` for targets")
        if not sequences.digital:
            raise ValueError("expected digital mode `SequenceFile` for targets")
        assert sequences.alphabet is not None
        alphabet = alphabet or sequences.alphabet
        targets: typing.Union["SequenceFile[DigitalSequence]", DigitalSequenceBlock] = sequences
    elif isinstance(sequences, DigitalSequenceBlock):
        alphabet = alphabet or sequences.alphabet
        targets = sequences
    else:
        queries = peekable(queries)
        try:
            alphabet = alphabet or queries.peek().alphabet or Alphabet.amino()
            targets = DigitalSequenceBlock(alphabet, sequences)
        except StopIteration:
            alphabet = alphabet or Alphabet.amino()
            targets = DigitalSequenceBlock(alphabet)

    # attempt to optimize parallelism based on the number of queries --
    # for few queries it's probably more efficient to parallelize on targets
    # instead, but we need to have a DigitalSequenceBlock for that
    _queries_hint = operator.length_hint(queries)
    _few_queries = _queries_hint != 0 and _queries_hint < cpus
    if parallel is None:
        if _few_queries and isinstance(targets, DigitalSequenceBlock):
            parallel = "targets"
        else:
            parallel = "queries"
    if parallel == "targets" and not isinstance(targets, DigitalSequenceBlock):
        raise RuntimeError("cannot use ``targets`` parallel mode with a sequence file")

    # start the dispatcher
    if "alphabet" not in options:
        options["alphabet"] = alphabet
    dclass = _SEARCHDispatcher if parallel == "queries" else _ReverseSEARCHDispatcher
    dispatcher = dclass(
        queries=queries,
        targets=targets,
        cpus=cpus,
        backend=backend,
        callback=callback,  # type: ignore
        builder=None,
        **options,
    )
    return dispatcher.run()  # type: ignore



import io
import itertools
import os
import shutil
import unittest
import tempfile

import pyhmmer
from pyhmmer.errors import EaselError
from pyhmmer.easel import SequenceFile
from pyhmmer.plan7 import HMMFile, HMMPressedFile, Pipeline

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files

# --- Mixins -------------------------------------------------------------------

class _TestHMMFile:

    ID = NotImplemented
    NAMES = NotImplemented

    @classmethod
    def setUpClass(cls):
        cls.hmms_folder = resource_files(__package__).joinpath("data", "hmms")

    def open_hmm(self, path):
        raise NotImplementedError()

    def check_hmmfile(self, hmmfile):
        for hmm, expected in itertools.zip_longest(hmmfile, self.NAMES):
            self.assertIsNot(hmm, None, "missing HMM: {}".format(expected))
            self.assertIsNot(expected, None, "unexpected extra HMM: {}".format(hmm))
            self.assertEqual(hmm.name, expected)
            self.assertIsNot(hmm.cutoffs, None)
            self.assertIsNot(hmm.evalue_parameters, None)

    def test_empty(self):
        with tempfile.NamedTemporaryFile() as empty:
            self.assertRaises(EOFError, self.open_hmm, empty.name)

    def test_read_hmmpressed(self):
        path = os.path.join(self.hmms_folder, "db", "{}.hmm".format(self.ID))
        with self.open_hmm(path) as f:
            self.check_hmmfile(f)

    def test_read_h3m(self):
        path = os.path.join(self.hmms_folder, "bin", "{}.h3m".format(self.ID))
        with self.open_hmm(path) as f:
            self.check_hmmfile(f)

    def test_read_hmm3(self):
        path = os.path.join(self.hmms_folder, "txt", "{}.hmm".format(self.ID))
        with self.open_hmm(path) as f:
            self.check_hmmfile(f)

    def test_read_hmm2(self):
        path = os.path.join(self.hmms_folder, "txt2", "{}.hmm2".format(self.ID))
        with self.open_hmm(path) as f:
            self.check_hmmfile(f)


class _TestHMMFileobj:

    def open_hmm(self, path):
        with open(path, "rb") as f:
            buffer = io.BytesIO(f.read())
        return HMMFile(buffer)

    def test_name(self):
        path = os.path.join(self.hmms_folder, "db", "{}.hmm".format(self.ID))
        with self.open_hmm(path) as f:
            self.assertIs(f.name, None)


class _TestHMMPath:

    def open_hmm(self, path):
        return HMMFile(path)

    def open_pressed(self, path):
        return HMMPressedFile(path)

    def test_init_error_filenotfound(self):
        self.assertRaises(FileNotFoundError, HMMFile, "path/to/missing/file")

    def test_init_error_folder(self):
        folder = tempfile.gettempdir()
        self.assertRaises(IsADirectoryError, HMMFile, folder)

    def test_read_optimized_profiles(self):
        path = os.path.join(self.hmms_folder, "db", "{}.hmm".format(self.ID))
        with self.open_hmm(path) as f:
            self.check_hmmfile(f.optimized_profiles())

    def test_optimized_profiles_length(self):
        path = os.path.join(self.hmms_folder, "db", "{}.hmm".format(self.ID))
        with self.open_hmm(path) as f:
            profiles = f.optimized_profiles()
            self.assertEqual(len(profiles), len(self.NAMES))
            profiles.read()
            self.assertEqual(len(profiles), len(self.NAMES) - 1)
            profiles.rewind()
            self.assertEqual(len(profiles), len(self.NAMES))

    def test_rewind(self):
        path = os.path.join(self.hmms_folder, "txt", "{}.hmm".format(self.ID))
        with self.open_hmm(path) as f:
            hmm1 = f.read()
            f.rewind()
            hmm2 = f.read()
            self.assertIsNot(hmm1, hmm2)
            self.assertEqual(hmm1.name, hmm2.name)

    def test_rewind_optimized_profiles(self):
        path = os.path.join(self.hmms_folder, "db", "{}.hmm".format(self.ID))
        with self.open_pressed(path) as f:
            om1 = f.read()
            f.rewind()
            om2 = f.read()
            self.assertIsNot(om1, om2)
            self.assertEqual(om1.name, om2.name)

    def test_name(self):
        path = os.path.join(self.hmms_folder, "db", "{}.hmm".format(self.ID))
        with self.open_hmm(path) as f:
            self.assertEqual(f.name, path)
            self.assertEqual(f.optimized_profiles().name, path)


class _TestThioesterase(_TestHMMFile):
    ID = "Thioesterase"
    NAMES = [b"Thioesterase"]


class _TestRREFam(_TestHMMFile):
    ID = "RREFam"
    NAMES = [
        b"Bottromycin_Methyltransferase_RRE",
        b"Cyanobactin_D_RRE",
        b"Cytolysin_C_RRE",
        b"Stand_Alone_Lasso_RRE",
        b"Lasso_Fused_RRE",
        b"Goadsporin_CD_RRE",
        b"Heterocycloanthracin_C_RRE",
        b"Lanthipeptide_LanB_RRE",
        b"Mycofactocin_RRE",
        b"NHLP_CD_RRE",
        b"Other_Sactipeptide_rSAM_RRE",
        b"Pantocin_Microcin_RRE",
        b"Plantazolicin_C_RRE",
        b"PlpY_RRE",
        b"PqqD_RRE",
        b"Proteusin_Methyltransferase_RRE",
        b"Proteusin_Epimerase_RRE",
        b"Quinohemoprotein_rSAM_RRE",
        b"Ranthipeptide_rSAM_RRE",
        b"Streptide_RRE",
        b"Subtilosin_rSAM_RRE",
        b"Thiaglutamate_B_RRE",
        b"Thiaglutamate_G_RRE",
        b"Thiaglutamate_I_RRE",
        b"Thiopeptide_F_RRE",
        b"Thuricin_rSAM_RRE",
        b"Thurincin_rSAM_RRE",
        b"Trifolitoxin_RRE"
    ]


# --- Test cases ---------------------------------------------------------------

@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestFileobjSingle(_TestHMMFileobj, _TestThioesterase, unittest.TestCase):
    pass

@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestFileObjMultiple(_TestHMMFileobj, _TestRREFam, unittest.TestCase):
    pass

@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestPathSingle(_TestHMMPath, _TestThioesterase, unittest.TestCase):
    pass

@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestPathMultiple(_TestHMMPath, _TestRREFam, unittest.TestCase):
    pass



import socket
import typing
import types

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

import pyhmmer.plan7
from pyhmmer.easel import Sequence, MSA
from pyhmmer.plan7 import TopHits, HMM, Builder

BIT_CUTOFFS = Literal["gathering", "trusted", "noise"]

S = typing.TypeVar("S", bound=Sequence)

class Client:
    address: str
    port: int
    socket: socket.socket
    def __init__(self, address: str = "127.0.0.1", port: int = 51371) -> None: ...
    def __enter__(self) -> Client: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def __repr__(self) -> str: ...
    def connect(self) -> None: ...
    def close(self) -> None: ...
    def search_seq(
        self,
        query: S,
        db: int = 1,
        ranges: typing.Optional[typing.List[typing.Tuple[int, int]]] = None,
        *,
        bias_filter: bool = True,
        null2: bool = True,
        seed: int = 42,
        Z: typing.Optional[float] = None,
        domZ: typing.Optional[float] = None,
        F1: float = 0.02,
        F2: float = 1e-3,
        F3: float = 1e-5,
        E: float = 10.0,
        T: typing.Optional[float] = None,
        domE: float = 10.0,
        domT: typing.Optional[float] = None,
        incE: float = 0.01,
        incT: typing.Optional[float] = None,
        incdomE: float = 0.01,
        incdomT: typing.Optional[float] = None,
        bit_cutoffs: typing.Optional[BIT_CUTOFFS] = None,
    ) -> TopHits[S]: ...
    def search_hmm(
        self,
        query: HMM,
        db: int = 1,
        ranges: typing.Optional[typing.List[typing.Tuple[int, int]]] = None,
        *,
        bias_filter: bool = True,
        null2: bool = True,
        seed: int = 42,
        Z: typing.Optional[float] = None,
        domZ: typing.Optional[float] = None,
        F1: float = 0.02,
        F2: float = 1e-3,
        F3: float = 1e-5,
        E: float = 10.0,
        T: typing.Optional[float] = None,
        domE: float = 10.0,
        domT: typing.Optional[float] = None,
        incE: float = 0.01,
        incT: typing.Optional[float] = None,
        incdomE: float = 0.01,
        incdomT: typing.Optional[float] = None,
        bit_cutoffs: typing.Optional[BIT_CUTOFFS] = None,
    ) -> TopHits[HMM]: ...
    def scan_seq(
        self,
        query: S,
        db: int = 1,
        *,
        bias_filter: bool = True,
        null2: bool = True,
        seed: int = 42,
        Z: typing.Optional[float] = None,
        domZ: typing.Optional[float] = None,
        F1: float = 0.02,
        F2: float = 1e-3,
        F3: float = 1e-5,
        E: float = 10.0,
        T: typing.Optional[float] = None,
        domE: float = 10.0,
        domT: typing.Optional[float] = None,
        incE: float = 0.01,
        incT: typing.Optional[float] = None,
        incdomE: float = 0.01,
        incdomT: typing.Optional[float] = None,
        bit_cutoffs: typing.Optional[BIT_CUTOFFS] = None,
    ) -> TopHits[S]: ...
    def iterate_seq(
        self,
        query: Sequence,
        db: int = 1,
        ranges: typing.Optional[typing.List[typing.Tuple[int, int]]] = None,
        builder: typing.Optional[Builder] = None,
        select_hits: typing.Optional[typing.Callable[[TopHits[HMM]], None]] = None,
        *,
        bias_filter: bool = True,
        null2: bool = True,
        seed: int = 42,
        Z: typing.Optional[float] = None,
        domZ: typing.Optional[float] = None,
        F1: float = 0.02,
        F2: float = 1e-3,
        F3: float = 1e-5,
        E: float = 10.0,
        T: typing.Optional[float] = None,
        domE: float = 10.0,
        domT: typing.Optional[float] = None,
        incE: float = 0.001,
        incT: typing.Optional[float] = None,
        incdomE: float = 0.001,
        incdomT: typing.Optional[float] = None,
        bit_cutoffs: typing.Optional[BIT_CUTOFFS] = None,
    ) -> IterativeSearch: ...
    def iterate_hmm(
        self,
        query: HMM,
        db: int = 1,
        ranges: typing.Optional[typing.List[typing.Tuple[int, int]]] = None,
        builder: typing.Optional[Builder] = None,
        select_hits: typing.Optional[typing.Callable[[TopHits[HMM]], None]] = None,
        *,
        bias_filter: bool = True,
        null2: bool = True,
        seed: int = 42,
        Z: typing.Optional[float] = None,
        domZ: typing.Optional[float] = None,
        F1: float = 0.02,
        F2: float = 1e-3,
        F3: float = 1e-5,
        E: float = 10.0,
        T: typing.Optional[float] = None,
        domE: float = 10.0,
        domT: typing.Optional[float] = None,
        incE: float = 0.001,
        incT: typing.Optional[float] = None,
        incdomE: float = 0.001,
        incdomT: typing.Optional[float] = None,
        bit_cutoffs: typing.Optional[BIT_CUTOFFS] = None,
    ) -> IterativeSearch: ...

class IterativeSearch(pyhmmer.plan7.IterativeSearch):
    client: Client
    db: int
    def __init__(
        self,
        client: Client,
        query: typing.Union[HMM, Sequence],
        db: int,
        builder: Builder,
        ranges: typing.Optional[typing.List[typing.Tuple[int, int]]] = None,
        select_hits: typing.Optional[typing.Callable[[TopHits[HMM]], None]] = None,
        options: typing.Optional[typing.Dict[str, object]] = None,
    ) -> None: ...



import abc
import copy
import gc
import io
import os
import pickle
import unittest
import tempfile
import warnings

from pyhmmer import easel
from pyhmmer.errors import AlphabetMismatch


class _TestSequenceBase(abc.ABC):

    @classmethod
    @abc.abstractmethod
    def Sequence(cls, **kwargs):
        pass

    def test_init_empty(self):
        seq = self.Sequence()
        self.assertEqual(seq.name, b"")
        self.assertEqual(len(seq), 0)
        self.assertEqual(len(seq.sequence), 0)

    def test_setter_name(self):
        seq = self.Sequence()
        self.assertEqual(seq.name, b"")
        seq.name = b"OTHER"
        self.assertEqual(seq.name, b"OTHER")

    def test_setter_description(self):
        seq = self.Sequence()
        self.assertIs(seq.description, b"")
        seq.description = b"a test sequence"
        self.assertEqual(seq.description, b"a test sequence")

    def test_setter_residue_markup(self):
        seq = self.Sequence(sequence="ATTGC")
        self.assertEqual(seq.residue_markups, {})
        seq.residue_markups = {b"kind": b"DADDC"}
        self.assertEqual(seq.residue_markups[b"kind"], b"DADDC")
        with self.assertRaises(ValueError):
            seq.residue_markups = {b"kind": b"D"}

    def test_pickle(self):
        s1 = self.Sequence(name=b"test1", sequence="GAATTC")
        s1a = pickle.loads(pickle.dumps(s1))
        self.assertEqual(s1.name, s1a.name)
        self.assertEqual(s1.sequence, s1a.sequence)
        self.assertEqual(s1.description, b"")
        self.assertEqual(s1.residue_markups, {})

        s2 = self.Sequence(
            name=b"test2", 
            sequence="TTCAAC", 
            residue_markups={b"structure": b"((.))."}
        )
        s2a = pickle.loads(pickle.dumps(s2))
        self.assertEqual(s2.name, s2a.name)
        self.assertEqual(s2.sequence, s2a.sequence)
        self.assertEqual(s2.residue_markups, s2a.residue_markups)

    def test_copy(self):
        seq = self.Sequence(name=b"TEST", sequence="ATGC")

        # copy the sequence
        cpy = seq.copy()

        # check the copy is equal and attributes are equal
        self.assertEqual(seq, cpy)
        self.assertEqual(seq.name, cpy.name)
        self.assertEqual(seq.checksum(), cpy.checksum())

        # check attributes can be read even after the original object
        # is (hopefully) deallocated, to make sure internal strings were copied
        del seq
        gc.collect()
        self.assertEqual(cpy.name, b"TEST")

        # check `copy.copy` works too
        cpy2 = copy.copy(cpy)
        self.assertEqual(cpy, cpy2)
        self.assertEqual(cpy.name, cpy2.name)

    def test_eq(self):
        seq1 = self.Sequence(name=b"TEST", sequence="ATGC")
        self.assertEqual(seq1, seq1)
        self.assertNotEqual(seq1, object())

        seq2 = self.Sequence(name=b"TEST", sequence="ATGC")
        self.assertEqual(seq1, seq2)

        seq3 = self.Sequence(name=b"OTHER", sequence="ATGC")
        self.assertNotEqual(seq1, seq3)

        seq4 = self.Sequence(name=b"TEST", sequence="ATGT")
        self.assertNotEqual(seq1, seq4)

    def test_eq_not_seq(self):
        seq1 = self.Sequence(name=b"TEST", sequence="ATGC")
        self.assertNotEqual(seq1, 1)
        self.assertNotEqual(seq1, b"hello")


class TestSequence(unittest.TestCase):

    def test_init_abstract(self):
        self.assertRaises(TypeError, easel.Sequence, name=b"ecoRI", sequence="GAATTC")

    def test_write_roundtrip(self):
        seq = easel.TextSequence(name=b"ecoRI", sequence="GAATTC")
        with io.BytesIO() as buffer:
            seq.write(buffer)
            seq2 = easel.SequenceFile.parse(buffer.getvalue(), "fasta")
        self.assertEqual(seq, seq2)


class TestDigitalSequence(_TestSequenceBase, unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.abc = easel.Alphabet.dna()

    @classmethod
    def Sequence(cls, **kwargs):
        seq = kwargs.get("sequence")
        if seq:
            kwargs["sequence"] = cls.abc.encode(seq)
        return easel.DigitalSequence(cls.abc, **kwargs)

    def test_sample(self):
        seq = easel.DigitalSequence.sample(self.abc, 10)
        self.assertLessEqual(len(seq), 10)
        self.assertIs(seq.alphabet, self.abc)

    def test_sample_seed(self):
        seq1 = easel.DigitalSequence.sample(self.abc, 10, randomness=easel.Randomness(42))
        seq2 = easel.DigitalSequence.sample(self.abc, 10, randomness=42)
        self.assertEqual(seq1, seq2)
        seq3 = easel.DigitalSequence.sample(self.abc, 10, randomness=100)
        self.assertNotEqual(seq1, seq3)

        with self.assertRaises(TypeError):
            seq4 = easel.DigitalSequence.sample(self.abc, 10, randomness="abc")

    def test_init_kwargs(self):
        arr = bytearray([0, 1, 2, 3])
        seq = easel.DigitalSequence(self.abc, name=b"TEST", sequence=arr)
        self.assertEqual(seq.name, b"TEST")
        self.assertEqual(bytearray(seq.sequence), arr)
        self.assertEqual(len(seq), 4)

    def test_textize_roundtrip(self):
        arr = bytearray([0, 1, 2, 3])
        dsq1 = easel.DigitalSequence(self.abc, name=b"TEST", sequence=arr)
        tsq1 = dsq1.textize()
        self.assertEqual(tsq1.name, dsq1.name)

        dsq2 = tsq1.digitize(self.abc)
        self.assertEqual(dsq1.name, dsq2.name)
        self.assertEqual(list(dsq1.sequence), list(dsq2.sequence))
        self.assertEqual(dsq1, dsq2)

    def test_reverse_complement(self):
        seq = easel.TextSequence(sequence="ATGC").digitize(self.abc)
        rc = seq.reverse_complement()
        self.assertEqual(rc.textize().sequence, "GCAT")

    def test_reverse_complement_inplace(self):
        seq = easel.TextSequence(sequence="ATGC").digitize(self.abc)
        seq.reverse_complement(inplace=True)
        self.assertEqual(seq.textize().sequence, "GCAT")

    def test_reverse_complement_protein(self):
        abc = easel.Alphabet.amino()
        seq = easel.TextSequence(sequence="MEMLP").digitize(abc)
        self.assertRaises(ValueError, seq.reverse_complement)
        self.assertRaises(ValueError, seq.reverse_complement, inplace=True)

    def test_invalid_characters(self):
        self.assertRaises(ValueError, easel.DigitalSequence, self.abc, name=b"TEST", sequence=b"test")

    def test_memoryview(self):
        seq = easel.TextSequence(sequence="ATGC").digitize(self.abc)
        mem = memoryview(seq)
        self.assertEqual(mem.tobytes(), b"\0\3\2\1")

    def test_translate(self):
        gencode = easel.GeneticCode()
        seq = easel.TextSequence(sequence="ATGCTGCCCGGTTTGGCACTGCTCCTGCTGGCCGCC").digitize(gencode.nucleotide_alphabet)
        prot = seq.translate(gencode).textize()
        self.assertEqual(prot.sequence, "MLPGLALLLLAA")

    def test_translate_copy_metadata(self):
        gencode = easel.GeneticCode()
        textseq = easel.TextSequence(sequence="ATGCTGCCCGGT", name=b"seq", source=b"source")
        seq = textseq.digitize(gencode.nucleotide_alphabet)
        prot = seq.translate(gencode).textize()
        self.assertEqual(prot.name, textseq.name)      
        self.assertEqual(prot.source, textseq.source)      

    def test_translate_empty(self):
        gencode = easel.GeneticCode()
        seq = easel.DigitalSequence(gencode.nucleotide_alphabet)
        prot = seq.translate(gencode).textize()
        self.assertEqual(prot.sequence, "")

    def test_translate_alphabet_mismatch(self):
        gencode = easel.GeneticCode()
        with self.assertRaises(AlphabetMismatch):
            prot = easel.DigitalSequence(gencode.amino_alphabet).translate(gencode)


class TestTextSequence(_TestSequenceBase, unittest.TestCase):

    @classmethod
    def Sequence(cls, **kwargs):
        return easel.TextSequence(**kwargs)

    def test_sample(self):
        seq = easel.TextSequence.sample(10)
        self.assertLessEqual(len(seq), 10)

    def test_sample_seed(self):
        seq1 = easel.TextSequence.sample(10, randomness=easel.Randomness(42))
        seq2 = easel.TextSequence.sample(10, randomness=42)
        self.assertEqual(seq1, seq2)
        seq3 = easel.TextSequence.sample(10, randomness=100)
        self.assertNotEqual(seq1, seq3)

        with self.assertRaises(TypeError):
            seq4 = easel.TextSequence.sample(10, randomness="abc")

    def test_init_kwargs(self):
        seq = easel.TextSequence(name=b"TEST", sequence="ATGC")
        self.assertEqual(seq.name, b"TEST")
        self.assertEqual(seq.sequence, "ATGC")
        self.assertEqual(len(seq), 4)

    def test_digitize_roundtrip(self):
        seq1 = easel.TextSequence(name=b"TEST", sequence="ATGC")
        dsq1 = seq1.digitize(easel.Alphabet.dna())
        self.assertEqual(seq1.name, dsq1.name)

        seq2 = dsq1.textize()
        self.assertEqual(seq1.name, seq2.name)
        self.assertEqual(seq1, seq2)

    def test_digitize_invalid(self):
        seq1 = easel.TextSequence(name=b"TEST", sequence=">v3ry 1nv4l1d")
        self.assertRaises(ValueError, seq1.digitize, easel.Alphabet.dna())
        self.assertRaises(ValueError, seq1.digitize, easel.Alphabet.amino())

    def test_memoryview(self):
        seq = self.Sequence(sequence="ATGC")
        mem = memoryview(seq)
        self.assertEqual(mem.tobytes(), b"ATGC")

    def test_reverse_complement(self):
        seq = easel.TextSequence(sequence="ATGC")
        rc = seq.reverse_complement()
        self.assertEqual(rc.sequence, "GCAT")

    def test_reverse_complement_inplace(self):
        seq = easel.TextSequence(sequence="ATGC")
        seq.reverse_complement(inplace=True)
        self.assertEqual(seq.sequence, "GCAT")

    def test_reverse_complement_protein(self):
        seq = easel.TextSequence(sequence="MEMLP")

        with warnings.catch_warnings():
            warnings.simplefilter('error')
            self.assertRaises(Warning, seq.reverse_complement)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            rc = seq.reverse_complement()
            self.assertEqual(rc.sequence, "NNKNK")




from libc.stdint cimport int64_t, uint32_t, uint64_t
from posix.types cimport off_t

from libeasel.msa cimport ESL_MSA

cdef extern from "esl_msaweight.h" nogil:


    ctypedef struct ESL_MSAWEIGHT_CFG:
        float fragthresh
        float symfrac
        bint ignore_rf
        bint allow_samp
        int sampthresh
        int nsamp
        int maxfrag
        uint64_t seed
        int filterpref

    const float eslMSAWEIGHT_FRAGTHRESH
    const float eslMSAWEIGHT_SYMFRAC
    const bint eslMSAWEIGHT_IGNORE_RF
    const bint eslMSAWEIGHT_ALLOW_SAMP
    const int eslMSAWEIGHT_SAMPTHRESH
    const int eslMSAWEIGHT_NSAMP
    const int eslMSAWEIGHT_MAXFRAG
    const int eslMSAWEIGHT_RNGSEED

    enum:
        eslMSAWEIGHT_FILT_CONSCOVER
        eslMSAWEIGHT_FILT_RANDOM
        eslMSAWEIGHT_FILT_ORIGORDER

    ctypedef struct ESL_MSAWEIGHT_DAT:
        uint64_t seed
        bint cons_by_rf
        bint cons_by_sample
        bint cons_by_all
        bint cons_allcos
        bint rejected_sample
        int ncons
        int* conscols
        int all_nfrag
        int samp_nfrag


    int esl_msaweight_PB(ESL_MSA *msa)
    int esl_msaweight_PB_adv(const ESL_MSAWEIGHT_CFG *cfg, ESL_MSA *msa, ESL_MSAWEIGHT_DAT *dat)

    ESL_MSAWEIGHT_CFG *esl_msaweight_cfg_Create()
    void               esl_msaweight_cfg_Destroy(ESL_MSAWEIGHT_CFG *cfg)
    ESL_MSAWEIGHT_DAT *esl_msaweight_dat_Create()
    int                esl_msaweight_dat_Reuse  (ESL_MSAWEIGHT_DAT *dat)
    void               esl_msaweight_dat_Destroy(ESL_MSAWEIGHT_DAT *dat)

    int esl_msaweight_GSC(ESL_MSA *msa)
    int esl_msaweight_BLOSUM(ESL_MSA *msa, double maxid)

    int esl_msaweight_IDFilter(const ESL_MSA *msa, double maxid, ESL_MSA **ret_newmsa)
    int esl_msaweight_IDFilter_adv(const ESL_MSAWEIGHT_CFG *cfg, const ESL_MSA *msa, double maxid, ESL_MSA **ret_newmsa)



#!/usr/bin/python

# Script to convert HMMER/Infernal/Easel test suite output into JUnit XML  for Teamcity integration
from lxml import etree
import sys


if len(sys.argv) != 3:
	print"Usage: parse_test_results.py <infile_name> <outfile_name>"
	quit()

infile_name = sys.argv[1]
outfile_name = sys.argv[2]

infile = open(infile_name, "r")  #open the input file

#root of the  XML tree is a "testsuites" entry
root = etree.Element("testsuites")

suite_passed = 0
suite_failed = 0
total_passed = 0
total_failed = 0
suite = "NONE"

#iterate over the lines in the file
for line in infile:
	tokens = line.split()

	#need to test length of tokens because there are some empty lines in the test output that
	#crash the script if you ask for their first element
	if len(tokens) > 0 and tokens[0] == "Running":
		#This is the start of a new test suite
		if suite != "NONE":
			#need to close out the previous suite before we start a new one
			suite.set("errors", "0")    #format expects these values even though we don't use them
			suite.set("skipped", "0")  #format expectes these values even though we don't use them
			suite.set("tests", str(suite_passed+suite_failed))
			suite.set("failures", str(suite_failed))
			suite.set("time", "0") # Just fake this since we don't track test time
			suite.set("timestamp", "2013-05-24T10:23:58")
			suite_passed = 0
			suite_failed = 0
		suitename = tokens[1].lower()
		suite = etree.SubElement(root, "testsuite")
		suite.set("name", suitename)

	if len(tokens) > 0 and tokens[0] == "exercise":
		#this is a new test result
		if(suite == "NONE"):
			print"Error: attempting to add test but didn't find a suite first"
			quit()

		test = etree.SubElement(suite, "testcase");
		testname = tokens[-3][:-1]
		if testname[0] == '[':
			testname = testname[1:]
		test.set("classname", suite.get("name"))
		test.set("name", testname)
		test.set("time", "0")
		if tokens[-1] == "FAILED":
			suite_failed += 1;
			total_failed += 1;
			#test failed, add appropriate fields to record
			failed = etree.SubElement(test, "failure")
			#extract what information we can from the output
			if tokens[6] == "[crash!]":
				failed.set("message", "crash")
				failed.text ="crash"
		else:
			suite_passed += 1
			total_passed += 1

	#lines with other first tokens aren't relevant to us	

if suite != "NONE":
	#need to close out the previous suite before we start a new one
	suite.set("errors", "0")    #format expects these values even though we don't use them
	suite.set("skipped", "0")  #format expectes these values even though we don't use them
	suite.set("tests", str(suite_passed+suite_failed))
	suite.set("failures", str(suite_failed))
	suite.set("time", "0") # Just fake this since we don't track test time
	suite.set("timestamp", "2013-05-24T10:23:58")		


# Done with processing, so write the XML to a file for import into Teamcity
xml = etree.ElementTree(root)
xml.write(outfile_name, pretty_print=True)




cdef extern from "esl_gev.h" nogil:

    double esl_gev_pdf    (double x, double mu, double lambda, double alpha);
    double esl_gev_logpdf (double x, double mu, double lambda, double alpha);
    double esl_gev_cdf    (double x, double mu, double lambda, double alpha);
    double esl_gev_logcdf (double x, double mu, double lambda, double alpha);
    double esl_gev_surv   (double x, double mu, double lambda, double alpha);
    double esl_gev_logsurv(double x, double mu, double lambda, double alpha);
    double esl_gev_invcdf (double p, double mu, double lambda, double alpha);

    double esl_gev_generic_pdf   (double x, void *params);
    double esl_gev_generic_cdf   (double x, void *params);
    double esl_gev_generic_surv  (double x, void *params);
    double esl_gev_generic_invcdf(double p, void *params);

    int    esl_gev_Plot(FILE *fp, double mu, double lambda, double alpha,
    			   double (*func)(double x, double mu, double lambda, double alpha),
    			   double xmin, double xmax, double xstep);



    double esl_gev_Sample(ESL_RANDOMNESS *r, double mu, double lambda, double alpha);

    int esl_gev_FitComplete(double *x, int n,
    			       double *ret_mu, double *ret_lambda,
    			       double *ret_alpha);
    int esl_gev_FitCensored(double *x, int n, int z, double phi,
    			       double *ret_mu, double *ret_lambda,
    			       double *ret_alpha);



import io
import itertools
import os
import shutil
import sys
import unittest
import tempfile

import pyhmmer
from pyhmmer.errors import EaselError
from pyhmmer.easel import SequenceFile, TextMSA, DigitalMSA
from pyhmmer.plan7 import HMMFile, Pipeline, TopHits

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestAlignment(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "PF02826.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm = hmm_file.read()
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "938293.PRJEB85.HG003687.faa")
        with SequenceFile(seqs_path, digital=True, alphabet=hmm.alphabet) as seqs_file:
            seqs = seqs_file.read_block()
        pipeline = Pipeline(alphabet=hmm.alphabet)
        cls._hits = pipeline.search_hmm(hmm, seqs)
    
    def setUp(self):
        self.hits = self._hits.copy()
        self.ali = self.hits[0].best_domain.alignment

    def test_str(self):
        rendered = str(self.ali)
        lines = rendered.splitlines()
        self.assertEqual(len(lines), 5)
        self.assertTrue(lines[1].strip().startswith(self.hits.query.name.decode()))
        self.assertTrue(lines[3].strip().startswith(self.hits[0].name.decode()))

    @unittest.skipIf(sys.implementation.name == "pypy", "`getsizeof` not supported on PyPY")
    def test_sizeof(self):
        self.assertGreater(sys.getsizeof(self.ali), 0)


import copy
import gc
import io
import os
import unittest
import tempfile

from pyhmmer import easel

from .. import __name__ as __package__
from .utils import resource_files


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestSSIReader(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.h3i = resource_files(__package__).joinpath("data", "hmms", "db", "Thioesterase.hmm.h3i")

    def test_init_error_filenotfound(self):
        self.assertRaises(
            FileNotFoundError, easel.SSIReader, "path/to/missing/file.ssi"
        )

    def test_init_error_wrongformat(self):
        self.assertRaises(ValueError, easel.SSIReader, __file__)

    def test_find_name(self):
        with easel.SSIReader(self.h3i) as reader:
            entry = reader.find_name(b"Thioesterase")
            self.assertEqual(entry.fd, 0)
            self.assertEqual(entry.record_offset, 0)

    def test_find_name_closed(self):
        with easel.SSIReader(self.h3i) as reader:
            reader.close()
            self.assertRaises(ValueError, reader.find_name, b"Thioesterase")

    def test_find_name_error(self):
        with easel.SSIReader(self.h3i) as reader:
            self.assertRaises(KeyError, reader.find_name, b"Ketosynthase")

    def test_file_info(self):
        with easel.SSIReader(self.h3i) as reader:
            info = reader.file_info(0)
            self.assertEqual(info.format, 0)
            self.assertEqual(info.name, "Thioesterase.hmm")

    def test_file_info_error(self):
        with easel.SSIReader(self.h3i) as reader:
            self.assertRaises(IndexError, reader.file_info, 1)

    def test_file_info_closed(self):
        with easel.SSIReader(self.h3i) as reader:
            reader.close()
            self.assertRaises(ValueError, reader.file_info, 0)


class TestSSIWriter(unittest.TestCase):
    def test_init_error_fileexists(self):
        with tempfile.NamedTemporaryFile() as tmp:
            self.assertRaises(
                FileExistsError, easel.SSIWriter, tmp.name, exclusive=True
            )



from libc.stdint cimport int64_t, uint64_t
from libc.stdlib cimport calloc, malloc, realloc, free

cimport libeasel
from libhmmer.p7_tophits cimport P7_TOPHITS

from capacity cimport new_capacity


ctypedef struct ID_LENGTH:
    int64_t id
    int64_t length


ctypedef struct ID_LENGTH_LIST:
    ID_LENGTH* id_lengths
    size_t     count
    size_t     size


cdef inline ID_LENGTH_LIST* idlen_list_init(size_t size) noexcept nogil:
    cdef ID_LENGTH_LIST* l = <ID_LENGTH_LIST*> malloc(sizeof(ID_LENGTH_LIST))
    if l != NULL:
        l.count = 0
        l.size = size
        l.id_lengths = <ID_LENGTH*> calloc(size, sizeof(ID_LENGTH))
        if l.id_lengths == NULL:
            free(l)
            l = NULL
    return l

cdef inline void idlen_list_destroy(ID_LENGTH_LIST* l) noexcept nogil:
    if l != NULL:
        if l.id_lengths != NULL:
            free(l.id_lengths)
        free(l)

cdef inline int idlen_list_add(ID_LENGTH_LIST* l, int64_t id, int64_t L) noexcept nogil:
    if l.count > 0 and l.id_lengths[l.count - 1].id == id:
        l.id_lengths[l.count - 1].length = L
    else:
        if l.count == l.size:
            l.size = new_capacity(l.size, l.count + 1)
            l.id_lengths = <ID_LENGTH*> realloc(l.id_lengths, l.size * sizeof(ID_LENGTH))
            if l.id_lengths == NULL:
                return libeasel.eslEMEM
        l.id_lengths[l.count].id = id
        l.id_lengths[l.count].length = L
        l.count += 1
    return libeasel.eslOK

cdef inline int idlen_list_assign(ID_LENGTH_LIST* l, P7_TOPHITS* th) noexcept nogil:
    cdef uint64_t i
    cdef int64_t  j = 0
    for i in range(th.N):
        while j < l.count and th.hit[i].seqidx != l.id_lengths[j].id:
            j += 1
        if j == l.count:
            return libeasel.eslENOTFOUND
        th.hit[i].dcl[0].ad.L = l.id_lengths[j].length
    return libeasel.eslOK

cdef inline int idlen_list_clear(ID_LENGTH_LIST* l) noexcept nogil:
    l.count = 0



# coding: utf-8
"""Custom exception handler for Easel, that avoids system exits.
"""

# --- C imports --------------------------------------------------------------

from cpython.exc cimport PyErr_Clear, PyErr_Occurred, PyErr_Fetch, PyErr_Restore
from cpython.ref cimport PyObject

cdef extern from "stdarg.h" nogil:
    ctypedef struct va_list:
        pass

cdef extern from "stdio.h" nogil:
    cdef int vsprintf(char* s, const char* format, va_list arg)


# --- Python imports ---------------------------------------------------------

from .errors import EaselError


# --- Error handler ----------------------------------------------------------

cdef object _recover_error():
    cdef PyObject*  type
    cdef PyObject*  value
    cdef PyObject*  traceback
    if PyErr_Occurred():
        PyErr_Fetch(&type, &value, &traceback)
        if isinstance(<object> value, Exception):
            return <object> value
        else:
            return (<object> type)(<object> value)
    else:
        return None

cdef void _reraise_error() except *:
    cdef object error = _recover_error()
    if error is not None:
        raise error

# We register a custom exception handler so that the program should not crash,
# but simply raise an exception, while chaining
cdef void _py_handler(int errcode, int use_errno, char* sourcefile, int sourceline, char* format, va_list argp) except * nogil:
    cdef char[2048] errbuf
    cdef int        errlength

    with gil:
        # recover the internal error, if any
        error = _recover_error()

        # get the error message if possible
        errlength = vsprintf(<char*> &errbuf, format, argp)
        if errlength > 0:
            message = errbuf[:errlength].decode('utf-8', errors="replace")
        else:
            message = None

        # raise the exception and hope it's collected
        raise EaselError(errcode, message) from error

libeasel.esl_exception_SetHandler(<libeasel.esl_exception_handler_f> _py_handler)



from libc.stdio cimport FILE


cdef extern from "esl_mininizer.h" nogil:

    const int eslMIN_MAXITER
    const float eslMIN_CG_RTOL
    const float eslMIN_CG_ATOL
    const int eslMIN_BRACK_MAXITER
    const float eslMIN_BRACK_STEP
    const float eslMIN_BRENT_RTOL
    const float eslMIN_BRENT_ATOL
    const float eslMIN_DERIV_STEP


    ctypedef struct ESL_MIN_CFG:
        int max_iterations
        double cg_rtol
        double cg_atol
        double brent_rtol
        double brent_atol
        int brack_maxiter
        double derive_step
        double* u
        int n


    ctypedef struct ESL_MIN_DAT:
        int niter
        double* fx
        int* brack_n
        double* brack_ax
        double* brack_bx
        double* brack_cx
        double* brack_fa
        double* brack_fb
        double* brack_fc
        int* brent_n
        double* brent_x
        int* nfunc


    int esl_min_ConjugateGradientDescent(ESL_MIN_CFG *cfg, double *x, int n,
					    double (*func)(double *, int, void *),
					    void (*dfunc)(double *, int, void *, double *),
					    void *prm, double *ret_fx, ESL_MIN_DAT *dat)

    ESL_MIN_CFG *esl_min_cfg_Create(int n)
    void         esl_min_cfg_Destroy(ESL_MIN_CFG *cfg)

    ESL_MIN_DAT *esl_min_dat_Create(ESL_MIN_CFG *cfg)
    void         esl_min_dat_Destroy(ESL_MIN_DAT *dat)
    int          esl_min_dat_Dump(FILE *fp, ESL_MIN_DAT *dat)



import io
import itertools
import os
import shutil
import unittest
import tempfile

import pyhmmer
from pyhmmer.errors import EaselError
from pyhmmer.easel import SequenceFile, TextMSA, DigitalMSA
from pyhmmer.plan7 import HMMFile, Pipeline, TopHits

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestHit(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "PF02826.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm = hmm_file.read()
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "938293.PRJEB85.HG003687.faa")
        with SequenceFile(seqs_path, digital=True, alphabet=hmm.alphabet) as seqs_file:
            seqs = seqs_file.read_block()
        pipeline = Pipeline(alphabet=hmm.alphabet)
        cls._hits = pipeline.search_hmm(hmm, seqs)
    
    def setUp(self):
        self.hits = self._hits.copy()

    def test_name_property(self):
        hits = self.hits.copy()
        hit = hits[-1]

        self.assertEqual(hit.name, b"938293.PRJEB85.HG003687_187")
        self.assertEqual(hits[-1].name, b"938293.PRJEB85.HG003687_187")
        self.assertEqual(self.hits[-1].name, b"938293.PRJEB85.HG003687_187")

        hits[-1].name = b"new name"

        self.assertEqual(hit.name, b"new name")
        self.assertEqual(hits[-1].name, b"new name")
        self.assertEqual(self.hits[-1].name, b"938293.PRJEB85.HG003687_187")

        with self.assertRaises(TypeError):
            hit.name = None

    def test_accession_property(self):
        hits = self.hits.copy()
        hit = hits[-1]

        self.assertIs(hit.accession, None)
        self.assertIs(hits[-1].accession, None)
        self.assertIs(self.hits[-1].accession, None)

        hits[-1].accession = b"NEW"

        self.assertEqual(hit.accession, b"NEW")
        self.assertEqual(hits[-1].accession, b"NEW")
        self.assertIs(self.hits[-1].accession, None)

        hits[-1].accession = None

        self.assertIs(hit.accession, None)
        self.assertIs(hits[-1].accession, None)
        self.assertIs(self.hits[-1].accession, None)

    def test_length_property(self):
        hits = self.hits.copy()
        hit = hits[-1]

        self.assertEqual(hit.length, 281)
        self.assertEqual(hits[-1].length, 281)
        self.assertEqual(self.hits[-1].length, 281)

    def test_description_property(self):
        hits = self.hits.copy()
        hit = hits[-1]

        self.assertTrue(hit.description.startswith(b'# 202177 # 203019 #'))
        self.assertTrue(hits[-1].description.startswith(b'# 202177 # 203019 #'))
        self.assertTrue(self.hits[-1].description.startswith(b'# 202177 # 203019 #'))

        hits[-1].description = None

        self.assertIs(hit.description, None)
        self.assertIs(hits[-1].description, None)
        self.assertTrue(self.hits[-1].description.startswith(b'# 202177 # 203019 #'))

        hits[-1].description = b"NEW"

        self.assertEqual(hit.description, b"NEW")
        self.assertEqual(hits[-1].description, b"NEW")
        self.assertTrue(self.hits[-1].description.startswith(b'# 202177 # 203019 #'))

    def test_manual_inclusion(self):
        hit = self.hits[-1]

        hit.included = hit.reported = hit.dropped = hit.duplicate = False
        self.assertFalse(hit.dropped)
        self.assertFalse(hit.duplicate)
        self.assertFalse(hit.reported)
        self.assertFalse(hit.included)

        hit.included = True
        self.assertFalse(hit.duplicate)
        self.assertFalse(hit.dropped)
        self.assertTrue(hit.included)
        self.assertTrue(hit.reported) # included hits are always reported

    def test_manual_drop(self):
        hit = self.hits[-1]

        hit.dropped = False
        hit.included = hit.reported = True
        self.assertFalse(hit.dropped)
        self.assertTrue(hit.reported)
        self.assertTrue(hit.included)

        hit.dropped = True
        self.assertTrue(hit.dropped)
        self.assertTrue(hit.reported) # dropped hits may be reported
        self.assertFalse(hit.included) # dropped hits are never included
        
        hit.dropped = False
        self.assertFalse(hit.dropped)
        self.assertTrue(hit.reported)
        self.assertFalse(hit.included)

    def test_manual_duplicate(self):
        hit = self.hits[-1]

        hit.duplicate = hit.dropped = False
        hit.included = hit.reported = True
        self.assertFalse(hit.duplicate)
        self.assertTrue(hit.reported)
        self.assertTrue(hit.included)

        hit.duplicate = True
        self.assertTrue(hit.duplicate)
        self.assertFalse(hit.reported) # duplicate hits are never reported
        self.assertFalse(hit.included) # duplicate hits are never included
        
        hit.duplicate = False
        self.assertFalse(hit.duplicate)
        self.assertFalse(hit.reported) # duplicate hits are never reported
        self.assertFalse(hit.included) # duplicate hits are never included
        



import abc
import unittest

from ...errors import AlphabetMismatch
from ...easel import Alphabet, DigitalSequence, Randomness
from ...plan7 import Background, HMM, Profile, OptimizedProfile, OptimizedProfileBlock


class TestOptimizedProfileBlock(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.rng = Randomness(seed=0)
        cls.alphabet = Alphabet.amino()
        cls.background = Background(cls.alphabet)

    @classmethod
    def _random_hmm(cls, name, M=100):
        hmm = HMM.sample(cls.alphabet, M, cls.rng)
        hmm.name = name
        return hmm

    @classmethod
    def _random_profile(cls, name, M=100):
        hmm = cls._random_hmm(name, M=M)
        profile = Profile(hmm.M, hmm.alphabet)
        profile.configure(hmm, cls.background, 200)
        return profile

    @classmethod
    def _random_optimized_profile(cls, name, M=100):
        return cls._random_profile(name, M=M).to_optimized()

    def test_alphabet_mismatch(self):
        om1 = self._random_optimized_profile(b"profile1")
        dna = Alphabet.dna()
        block = OptimizedProfileBlock(dna)
        self.assertRaises(AlphabetMismatch, block.append, om1)
        self.assertRaises(AlphabetMismatch, block.insert, 0, om1)

    def test_truth(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")

        self.assertTrue(not OptimizedProfileBlock(self.alphabet))
        block = OptimizedProfileBlock(self.alphabet, [om1, om2])
        self.assertTrue(block)

    def test_identity(self):
        self.assertIsNot(OptimizedProfileBlock(self.alphabet), OptimizedProfileBlock(self.alphabet))
    
    def test_len(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        self.assertEqual(len(OptimizedProfileBlock(self.alphabet, [])), 0)
        self.assertEqual(len(OptimizedProfileBlock(self.alphabet, [om1])), 1)
        self.assertEqual(len(OptimizedProfileBlock(self.alphabet, [om1, om2, om3])), 3)

    def test_append(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")

        block = OptimizedProfileBlock(self.alphabet)
        self.assertEqual(len(block), 0)

        block.append(om1)
        self.assertEqual(len(block), 1)
        self.assertEqual(block[0], om1)

        block.append(om2)
        self.assertEqual(len(block), 2)
        self.assertEqual(block[1], om2)

    def test_iter(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2, om3])
        self.assertEqual(len(block), 3)

        iterator = iter(block)
        self.assertIs(next(iterator), om1)
        self.assertIs(next(iterator), om2)
        self.assertIs(next(iterator), om3)
        self.assertRaises(StopIteration, next, iterator)

    def test_clear(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2])
        self.assertEqual(len(block), 2)
        block.clear()
        self.assertEqual(len(block), 0)
        block.clear()
        self.assertEqual(len(block), 0)
        
        block = OptimizedProfileBlock(self.alphabet)
        self.assertEqual(len(block), 0)
        block.clear()
        self.assertEqual(len(block), 0)

    def test_getitem(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2])
        self.assertEqual(len(block), 2)

        self.assertIs(block[0], om1)
        self.assertIs(block[1], om2)
        self.assertIs(block[-1], om2)
        self.assertIs(block[-2], om1)

        with self.assertRaises(IndexError):
            block[3]
        with self.assertRaises(IndexError):
            block[-5]
        
    def test_setitem(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2])
        self.assertEqual(len(block), 2)

        block[0] = om3
        self.assertIs(block[0], om3)
        self.assertIs(block[1], om2)

    def test_remove(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2])
        block.remove(om1)
        self.assertEqual(len(block), 1)
        self.assertIs(block[0], om2)

    def test_index(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2])
        self.assertEqual(len(block), 2)

        self.assertEqual(block.index(om1), 0)
        self.assertEqual(block.index(om2), 1)

        self.assertRaises(ValueError, block.index, om3)
        self.assertRaises(ValueError, block.index, om1, start=1)

    def test_pop(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2, om3])
        self.assertEqual(len(block), 3)

        self.assertIs(block.pop(), om3)
        self.assertIs(block.pop(0), om1)

        self.assertEqual(len(block), 1)
        self.assertIs(block[0], om2)

        self.assertIs(block.pop(-1), om2)
        self.assertEqual(len(block), 0)
        self.assertRaises(IndexError, block.pop)

        empty = OptimizedProfileBlock(self.alphabet)
        self.assertRaises(IndexError, block.pop)

    def test_setitem_slice(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2])
        self.assertEqual(len(block), 2)
        self.assertIs(block[0], om1)
        self.assertIs(block[1], om2)

        block[:] = [om1, om2, om3]
        self.assertEqual(len(block), 3)
        self.assertIs(block[0], om1)
        self.assertIs(block[1], om2)
        self.assertIs(block[2], om3)

        block[2:5] = [om3, om3, om3]
        self.assertEqual(len(block), 5)
        self.assertIs(block[3], om3)
        self.assertIs(block[4], om3)

    def test_delitem_slice(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2, om3])
        self.assertEqual(len(block), 3)
        del block[:]
        self.assertEqual(len(block), 0)

        block = OptimizedProfileBlock(self.alphabet, [om1, om2, om3])
        self.assertEqual(len(block), 3)
        del block[::2]
        self.assertEqual(len(block), 1)
        self.assertIs(block[0], om2)

    def test_contains(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2])
        self.assertEqual(len(block), 2)
        self.assertIn(om1, block)
        self.assertIn(om2, block)
        self.assertNotIn(om3, block)
        
        self.assertNotIn(42, block)
        self.assertNotIn(object(), block)

    def test_copy(self):
        om1 = self._random_optimized_profile(b"profile1")
        om2 = self._random_optimized_profile(b"profile2")
        om3 = self._random_optimized_profile(b"profile3")

        block = OptimizedProfileBlock(self.alphabet, [om1, om2, om3])
        block_copy = block.copy()

        self.assertEqual(len(block), len(block_copy))
        self.assertEqual(block, block_copy)

        self.assertIs(block[0], block_copy[0])
        self.assertIs(block[1], block_copy[1])
        self.assertIs(block[2], block_copy[2])

        block.remove(om3)
        self.assertEqual(len(block), 2)
        self.assertEqual(len(block_copy), 3)


#! /usr/bin/env python3

# iss #159 : nhmmer reports overlapping envelopes on reverse strand
#
# A bug in how nhmmer sorts hits on reverse strand led to it reporting
# overlapping envelopes in some situations.
#
# [xref SRE:2019/0531-nhmmer-iss159]

import sys
import os
import subprocess

if len(sys.argv) != 4:
    sys.exit("Usage: issxxx-nhmmer-overlap.py <builddir> <srcdir> <tmppfx>")

builddir = sys.argv[1]
srcdir   = sys.argv[2]
tmppfx   = sys.argv[3]

if (not os.path.exists('{}/src/hmmbuild'.format(builddir))): sys.exit("FAIL: didn't find {}/src/hmmbuild".format(builddir))
if (not os.path.exists('{}/src/nhmmer'.format(builddir))):   sys.exit("FAIL: didn't find {}/src/nhmmer".format(builddir))

with open('{0}.sto'.format(tmppfx), 'w') as f:
    print ("""\
# STOCKHOLM 1.0
D1  TCCCAAATTCAGTCAAGATTCTTACGTGGAATTTCAGGAAAAAACTATTT.CGGGAGAATTTTTCAGAAATTTTTTGCAAACCAAGGCATCCTCTACCAAAACTTTTGAAATTTTGCCCTA...........................CTTTCACACACAAACGAACCTTCCTATCAGCCTGGGCAGTGATGAGTAAAAAACTTATTGCGAATCTTGTGGGACGAATCTGGAGCTCAAATCTAAGCCAAAACTCAATAGACACGATGACGAATGTCTGGTAAAAATTTCGGACCAAAA.TACCCAAGGAGTAAGGCGTAATAAGTCGCAGACCGGGAGTGAACAAAACCAGTTTTCCGAAAACAAAACGTCCTGGCTTTTCTAACCCATATCTTCTCTTTGTGATCTGTTTATGGACGTGCCTCATTTCTTGGGTGCAAACAACGTATGAAAGTACATGTATGAATTTTTT.CAGCGCAATACAGACCCAGAATAAAATTGCAGAAAGTAGGGTGAAATTTCACAAGTTTTGCTAGAGGATAGCTTT.GGTTTTCAGAAATTTTCTGAAAAATTT
H12 ...............................TCCCAAAATAGTTTTTTTC.GTGAAATTCCATAAAAGAATATCCATTGAATTTGGGACAAAACACGACAAATTTTAGGTCAAACGGATGAGTACTCACCCACGA...AAAAAATCAAACAGTACACACAAACAAACCTTCCTTTCAGTCCTGGCAGTGACGAATAAAAAATTTATTGTCAATCTTGTGACACTAATCTGGGGCTCAAATCTAAGCCAAAACTCAAAAGACACAGTGACGAATGTCTGGTAAAAATTTCATCCAAAAA.TACCTAAGGAGTAAGGCGGAGTAAATCCTAGACCGAGAGTGAACAAAACTGGTTTTCCGAAAACGAAACGTCCTGACTTGTTTTCCCCGTATCATCTATTTGTGATTTGTTTATGGACGTGTCTCCTTGCCTGGGTGCCAACAACATATTAAAGTACTTGTGCGAATTTTTT.CAGCGCTATACAGGCCTAGAATAAAATT.TTGAGAGTAGGGCGAAATTTCACAAGTTTTGGTAGAGGATAACCTT.GGTTTGCACAAAATTTATGAAGAATTC
H21 ...............................TCCCAAAATAGTTTTTTTC.CGAAAATTCCATGAAAGAATATCCATTGAATTTGGGACA..................................................................CACAAACGAACCTTCCTTTCAGTCCTGGCAGTGATGAATAAAAAATTTATTGTCAATCTTGTGACACTAATCTGGG.CTCAAATCTCAGCCAAAACTCAAAAGATACAGTGACGAATGTCTGGTAAAAATTTCATCCAAAAAATACCCAAGGATTAAGGCGGAGTAAATCCTAGACCGAGAGTGAACAAAACTGGTTTTCCGAAAACGAAACGTCCTGACTTGTTTTCCCCGTATCATCTATTTGTGATTTGTTTATGGACGTGCCTCCTTGCTTGGGTGCCAACAACATATTAAAGTACTTGTGCAAATTTTTT.C.....TATACAGGCCCAGAATAAAATT.CTGAGAGTAGGGCGAAATTTCACAAGTTTTGGTAGAGGATAACCTT.GGTTTGCACAAAATTTCTGAAGAATTC
L3  ...............................TCTCGAAATAGTTTTTTTTTCTGAAATTCCACGTAAGAATCTGCACTGAATTTGGGACAAAACGCGACAAATTTCGGGTTAAACGGATGAGTATTCACCCACGAAAAGAATCAAACAGTTTCACACACAAACGAAGCTTCCTTTCAGCCATGGCACTGATGAATAAAAAATTTATTGCCAATCTTGTGAGACGAATCTGGAGCTCAAGTCTCAGCCAAATCTCAATAGACACAGTGACGAATTTATGGTAAAAATTTCAGACCAAAA.TACCCAAGGAGTAAGGCGTAGTAAGTCCCAGAACGAGAGTGAACAAACCTAGTTTTCCGAAAACAAAACGTCCTGGCTTTTCTTCCCCGTATCTTACTTCTGTGATTTGTTTATGGACGTGCCTCCTTACCTGGGTGCAAACAACATATGAAAGTACTTGTGTGAATTTTTT.CAGCACAATACAGACCGAGAATAAAATTGCAAAAAGTTGGGCGAATTTTCACCAATTTTGCTAGAGCATAGCCTTTGGTTTGCACAAAATTTCTGAAAAATTC
L5  ...............................TCCCGAAATAGTTTTTTTC.CTGAAATTCCACGTAAGAATCTCCACTGAATTTGGGACAAAACGCGACAAATTTTGGGTTAAACGGATGAGTATTCACCCACGAAAAGAATCAAACAGTTTCACACACAAACGAAGCTTCCTTTCAGCCATGGTACTGATGAATAAAAAATTTATTGCCAATCTTGTGAGACGAATCTGGAGCTCAAGTCTCAGCCAAAACTCAATAGACATAGTGACGAATGTATGGTAAAAATTTCAGACCAAAA.TACCCAAGCAGAAAGGCGTAGTAAGTCCAGGACCGAGAGTGAACAAAACTGGTTTTCCGAAAACAAAACGTCATGGCTTTTCTTCCCCGTATCTTCCTTCTGTGATTTGTTTATGGACGTGCCTCCTTACCTGGATGCAAACAACATATGAAAGTACTTGTGTGAATTTTTT.CAGCGCAGTACAGACCAAGAATAAAATTGCACAAAGTTGGGCGAAATTTCACCAATTTTGGTAGAGGAT..................................
L1  ...............................TCCCGAAATAGTTTTTTTC.CTGAAATTCCTCATAAGAATCTCCACTTAGTTTGGGATAAAACGCGAAAAATTTCGAGTCAAACGGATGAGTATTCACCCACGAAAAAAATGAAACAGTTTCACACGCGAACGAACCTTCCTTTCAGCCCTGGCAGTGATGAATAAAAAATTTATTGCCAATGTTGTGGGACGAATCTGGAGCTCAAGTCTCAGCCAAAACTCAATAGACACAGTGACGAACGTGTGGTAAAAATTTTAGACCAAAA.TAGCCAAGGAGTAAGGCGTAGTAAGTCCCAGACCGAGAGTGAACAAATATGGTTTTCCGAAAACAAAACGTCCTGGCTTTTCTTCTCCGTATCTTCCTTCTGTGATTTTTTTATGGACGTGCCTCCTTACCTGGGTGCAAACAACATATGAAAGTACTTGTGTGAATTTTTT.CAGCGCAATACGGACCAAGAATAAAATTGCAGAAAATTGGGCGACATTTCACAAGTTTTGGTAGAGGATAGCCTTTGATTTGCACAAAATTTCTGAAAAATTC
L2  ...............................TCCCGAAATAGTCTTCTCC..TGAAATTCCACGTAACAATCTCCACTGAATTTGGGACAAAACGCGATAAATTTCTGGCCAAACGGCTGAGTATTCACCCACGAAAAGAATCACATAGTTTCACACACAAACGAACCTTCCTTTCAGCAATGGCAGTGATTAATAAAAAATTTATTGACAATCTCTTGGGACGAATCTGGAGCTCAAGTCTCAGCAAAAACTCAATAGACACAGTGACGAACGTCTGGTAAAAATTTCAGACCAAAA.TACCCAAGGAGTAAGGCGTAATAAGTCCCAGACCGAGAGTGAACAAAATTGGTTTTCCGAAAACAAAACGTCCAAGCTTTTCTTCCCCGTATCTTCCTTTTGTGATTTGTTTATGGACGTGCCGCCTTACCTGGGTGCAAAAAACATGTGAAAGTACTTGTAAGAATTTTTTTCAGTGCAATACAGACCCATAACAAAATTGCAGAAAGTTCAGCGAAATTTCACAAGTTTTGGTAGAGGATAGCCTTTAGTTTGCACAAAATTTCTGAAAAAATC
F5  ................................TTATGCTTTTGTTTCTA.CCTTA..TTTCTTCCAAGAATCTTCACTGAATTTGGGACAAAATGCGACAAATTTCAGGTCAAACGGATGAGTACTCACCCACGAAAAAAATCAAACAGTTTCACACACAAACGAACCTTCCTTTTAGCCCTGTCAGTAATGAATAAAAAATTCATTTCCAATCGTTTTGGAGGAATCTGGTGCTCAAATCTCAGCCAAAACTCAATAGACTCAGTGACGAATGTCTGGTAATAATATCAGGCAAAAA.TACCCAAGGAGTAAGGCATAGTAAGTCCCAGACCTAGAGTGAACAAAACAGGTTTTTGGAAAACGAAACGTCCTGGCTTCTCTTCCCCGTATCTTCTTTTTGTGATTTGTTTTTGGACGTGCCTCCTTACCTAGGTGCAAAAAACATATGAAAGTACTTTTGTAAATTTTTT.CAGTGCTATACGGGCCCAGAATAAAATTACAGAGAGTAGGGCGAAATTTCACAAGTTTTGGTAGAGGATAGTCTT.GGTTTGCACAAAATTTCTGAAAAATTC
H11 ................................ATCCGAAATAGTTTTTT.CCTGAAATTTCATGAGAGAATATCCACTGAATTTGGAAAAAAAGGCGACAAATTTCAGGTCAAACGGATGTGTACTCACCCACGAAAAAAATCAAACGGTTTCACACACAAGCAAACCTTCCTTTCAGCCCTGGCAGTGATGAATAAAAAGTTTATTGCCAATCTTCTGGCACGAATCTGGGGCTCAAATCTCAGTCAAAACTCAATAGACACAGTGGCAAATGTCTAGTAAAAATTTCAGACAAAAA.TACCGTATGAGTAAGGCGAAGTAAGTCCCATACCAAGAGTGAACAAAATTGATTTTCCAAAAAGGAAACGTCATGGCTTTTCTTCCCGGTATCTTCTTTTTATGATTTGTTTATGGTCGTGCCTCCTTGCCTGGGTGAAAACAACATACGAAAGTACTTGTGTGAATTTTTT.CAGCGCTATACA........................TAGGGCGAAATTTCACAAGTTTTGGTAGAG.ATAACCTT.GGTTTGCATAAAATTTCTGAAAAATTC
//""", file=f)

with open('{0}.fa'.format(tmppfx), 'w') as f:
    print("""
>bug_example
TATTACGGGAGAATTTTTCAGAAATTTTGTGCAAACCAAGGTTATCCTCTACCGAAACTTGTGAAATTCCGGCCTAATTTCTGCAATTTCATTCTGGGTCCGTATTGCGCTGAAAAAAATTCACACAGGCACTTTTGTATGTTGTTTGCACCCGGGCAAGGAGGCACGTCCATAAACGAATCACAAAAAGAATATACGGGGAAGAAAAGCCAGGACATTTTGTTTTCGGAAAACCGGTTT
AGTTCACTCTCGGACTGGGACTTACTACGCCTTACTCCTTGGGTATTTTGGTCTGAAATTTTTACCAGACATTCGTCACTGTGTCTATTGAGTTTTGGTTGTGGTTTGAACCCCAGGTTCGTCCCACAAGATTGGCAATAAATCTTTTATTCGTCACCGCCAGGGCTGAAAGGAAGGTTCGTTTGTGTGTGAAACTGTTTGATTTTTTTCGTGGGTGAATACTTATCCGTTTGACCTAAA
ATTTGTCGCGTTTTGTCCAAAATCCAATGGAGATTCTTGGGTAAAATTTCAGAAAAAAACTATTCCGGGAGAATTTTTCAGAAATTTTGTGCAAACCAAGGGTATCCTTTATTGAAACTTGTGAAATTCCGGCCTAATTTCTGCAATTTCATTCTGGGTCCGTATTGCACTGAAAAAAAATTCACACAAGCACTGTCGTATGTTGTTTGCACCTAGGCAGGGAGGCACGTCCATAAACGA
ATCACAAAAAGAAGATACGGGGAAGAAAAGCCAGGACGTTTTGTTTTCGGAAAACCGGTTTTGTTCCCTCTCGGTCTGGAACTCACTACGCCTTACTCCTTGGGTATTTTGGTATGAAATTTTTACCAGACATTCGACACTGTGTCAATTGAGTTTTGGATGAGGCTTGAGCCCCAGGTTCGGCCCACAAGATTGACAATAAATCTTTTAATCGTCAACGCCAGGGCTGAGAGGAAGGTT
CGTTTGTGTGTGAAACTGTTTTGATTTTTTTCGTGGGTGAATACTCATCCGTTTGACCTAAAATTTAACGCGTTTTGTCCCAAATTTAGTGGAGATTCTTGGGTGGAATTTCAGGAAAAAACTATATTGGGAGAATTTTTCAGAAATTTTGTTCAAACCAAGGCTATCCTCTACCAAAACTTATGAAATTTCGTCCTAATTTCTGCAATTTAATTATGGGTCCGTATTGCGTTGAAAAAA
ATTCACACAAGCACTTTCGTATATTTTTTGCACTCAAGCAAGGAGGCACGTCCATAAACGAATCACAAAAAGAAGATACGGGGAAGAAAAACCATAACGTTTTGTTTTCGGAAAACCGGTTTTGTTCACTCTCGGTCTGGGACTTACTACGACTTACTCCTTGGGTATTTTGGTCTGAAATTTTTACCAGACATTCGACAATGTGTCTATTGAGTTTCGGCTGAGGCTTGAGCCCCAGGT
TCGGCCCACAAGAGTGGCAATAAATCTTTTATTCGTCAATGCCACGGCTGAAAGGAAGGT""", file=f) 

# Make sure previous tests haven't left an indexed <tmppfx>.hmm file around, otherwise
# those index files + this about-to-be-created .hmm file will cause errors.
#
for sfx in [ 'h3f', 'h3i', 'h3m', 'h3p' ]:
    idxfile = '{0}.hmm.{1}'.format(tmppfx, sfx)
    if os.path.exists(idxfile):
        os.remove(idxfile)

try:
    retcode = subprocess.call(['{}/src/hmmbuild'.format(builddir), '{}.hmm'.format(tmppfx), '{}.sto'.format(tmppfx)], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
except:
    sys.exit("FAIL: hmmbuild failed")
if retcode != 0: sys.exit("FAIL: hmmbuild failed")

try:
    retcode = subprocess.call(['{}/src/nhmmer'.format(builddir), '--crick', '--tblout', '{}.tbl'.format(tmppfx), '{}.hmm'.format(tmppfx), '{}.fa'.format(tmppfx)], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
except:
    sys.exit("FAIL: nhmmer failed")
if retcode != 0: sys.exit("FAIL: nhmmer failed")


# From the .tbl output, extract a list of envelopes;
# each envelope defined by (start,end) tuple, with start <= end
# nhmmer is allowed to report overlapping hits on different strands, not same strand.
# The nhmmer run above used --crick to search only one strand (in this case, the reverse strand).
#
# The bug appears as an extra (551..792) envelope:
#
#  # target name        accession  query name           accession  hmmfrom hmm to alifrom  ali to envfrom  env to  sq len strand   E-value  score  bias  description of target
#  #------------------- ---------- -------------------- ---------- ------- ------- ------- ------- ------- ------- ------- ------ --------- ------ ----- ---------------------
#  bug_example          -          tmpfoo               -                2     540     549      11     550      11    1500    -    6.8e-176  574.5  14.7  -
#  bug_example          -          tmpfoo               -                4     540    1089     551    1092     551    1500    -    5.9e-170  554.9  13.7  -
#  bug_example          -          tmpfoo               -              134     540    1500    1093    1500    1093    1500    -      3e-135  440.3  11.1  -
#  bug_example          -          tmpfoo               -              301     540     792     551     792     551    1500    -     4.1e-76  245.1   5.2  -
#
envlist = []
with open('{0}.tbl'.format(tmppfx)) as f:
    for line in f:
        if line[0] == '#': continue   # skip comment lines
        fields = line.split()
        s      = int(fields[8])
        e      = int(fields[9])

        if s < e: envlist.append( ( s,e ))
        else:     envlist.append( ( e,s ))

# Detect overlapping envelopes: sort by start position, then look for
# any case where the start of an envelope is <= the end of the previous
# envelope.
envlist.sort(key=lambda elem: elem[0])
for i in range(1, len(envlist)):
    if envlist[i][0] <= envlist[i-1][1]: sys.exit("FAIL: overlapping envelopes detected")



os.remove('{0}.tbl'.format(tmppfx))
os.remove('{0}.hmm'.format(tmppfx))
os.remove('{0}.sto'.format(tmppfx))
os.remove('{0}.fa'.format(tmppfx))

print("ok")

        
 



import copy
import gc
import io
import os
import unittest
import tempfile
from itertools import zip_longest

from pyhmmer import easel

from ..utils import EASEL_FOLDER


class TestMSAFile(unittest.TestCase):

    def test_guess_alphabet_empty_sequence(self):
        buffer = io.BytesIO(b">seq1\n\n")
        self.assertRaises(ValueError, easel.MSAFile, buffer, format="afa", digital=True)

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_init_error_unknownformat(self):
        stockholm = os.path.join(EASEL_FOLDER, "formats", "stockholm.1")
        self.assertRaises(ValueError, easel.MSAFile, stockholm, format="nonsense")

    def test_init_error_empty(self):
        with tempfile.NamedTemporaryFile() as f:
            self.assertRaises(ValueError, easel.MSAFile, f.name)  # cannot guess format

    def test_init_file_not_found(self):
        self.assertRaises(FileNotFoundError, easel.MSAFile, "path/to/missing/file.sto")

    def test_init_error_folder(self):
        folder = tempfile.gettempdir()
        self.assertRaises(IsADirectoryError, easel.MSAFile, folder)

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_author(self):
        trna_5 = os.path.join(EASEL_FOLDER, "testsuite", "trna-5.stk")
        with easel.MSAFile(trna_5) as f:
            msa = f.read()
        self.assertEqual(msa.author, b"Infernal 0.1")

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_read_wrong_format(self):
        stockholm = os.path.join(EASEL_FOLDER, "formats", "stockholm.1")
        # check reading while giving another file format fails
        with easel.MSAFile(stockholm, "clustal") as f:
            self.assertRaises(ValueError, f.read)

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_iter(self):
        trna_5 = os.path.join(EASEL_FOLDER, "testsuite", "trna-5.stk")
        with easel.MSAFile(trna_5) as f:
            msa = next(f)
            self.assertRaises(StopIteration, next, f)

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_closed_file(self):
        trna_5 = os.path.join(EASEL_FOLDER, "testsuite", "trna-5.stk")
        with easel.MSAFile(trna_5) as f:
            pass
        self.assertRaises(ValueError, f.read)

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_reference(self):
        selex = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "selex", "selex.good.4")
        with easel.MSAFile(selex, format="selex") as f:
            msa = f.read()
        self.assertEqual(msa.reference, b"......xxxxxxx xxxx xxxxxx......")
        self.assertIsNone(msa.model_mask)

    @unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
    def test_secondary_structure(self):
        selex = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "selex", "selex.good.4")
        with easel.MSAFile(selex, format="selex") as f:
            msa = f.read()
        self.assertEqual(msa.secondary_structure, b"......>>>>+>> ^^^^ <<<<<<......")


class _TestReadFilename(object):

    def test_read_filename_guess_format(self):
        for filename, start, count, length in zip_longest(self.filenames, self.starts, self.counts, self.lengths):
            path = os.path.join(self.folder, filename)
            with easel.MSAFile(path) as f:
                msa = f.read()
                self.assertEqual(msa.sequences[0].sequence[:10], start)
                self.assertEqual(len(msa), length)
                self.assertEqual(len(msa.sequences), count)

    def test_read_filename_given_format(self):
        for filename, start, count, length in zip_longest(self.filenames, self.starts, self.counts, self.lengths):
            path = os.path.join(self.folder, filename)
            with easel.MSAFile(path, self.format) as f:
                msa = f.read()
                self.assertEqual(msa.sequences[0].sequence[:10], start)
                self.assertEqual(len(msa), length)
                self.assertEqual(len(msa.sequences), count)

    def test_guess_alphabet(self):
        for filename, alphabet in zip_longest(self.filenames, self.alphabet):
            path = os.path.join(self.folder, filename)
            with easel.MSAFile(path, self.format, digital=True) as f:
                self.assertEqual(f.alphabet, alphabet)


class _TestReadFileObject(object):

    def test_read_fileobject_guess_format(self):
        for filename, start, count, length in zip_longest(self.filenames, self.starts, self.counts, self.lengths):
            path = os.path.join(self.folder, filename)
            with open(path, "rb") as f:
                buffer = io.BytesIO(f.read())
            with easel.MSAFile(buffer) as f:
                msa = f.read()
                self.assertEqual(msa.sequences[0].sequence[:10], start)
                self.assertEqual(len(msa), length)
                self.assertEqual(len(msa.sequences), count)

    def test_read_filename_given_format(self):
        for filename, start, count, length in zip_longest(self.filenames, self.starts, self.counts, self.lengths):
            path = os.path.join(self.folder, filename)
            with open(path, "rb") as f:
                buffer = io.BytesIO(f.read())
            with easel.MSAFile(buffer, self.format) as f:
                msa = f.read()
                self.assertEqual(msa.sequences[0].sequence[:10], start)
                self.assertEqual(len(msa), length)
                self.assertEqual(len(msa.sequences), count)

    def test_guess_alphabet(self):
        for filename, alphabet in zip_longest(self.filenames, self.alphabet):
            path = os.path.join(self.folder, filename)
            with open(path, "rb") as f:
                buffer = io.BytesIO(f.read())
            with easel.MSAFile(buffer, self.format) as f:
                self.assertEqual(f.alphabet, alphabet)


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestA2MFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "a2m")
    filenames = ["a2m.good.1", "a2m.good.2"]
    starts    = ["VLSEGEWQLV", "UCCGAUAUAG"]
    format    = "a2m"
    lengths   = [165, 73]
    counts    = [4, 5]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.rna()]

    @unittest.expectedFailure
    def test_read_filename_guess_format(self):
        super().test_read_filename_guess_format()

    @unittest.expectedFailure
    def test_read_fileobject_guess_format(self):
        super().test_read_fileobject_guess_format()


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestAlignedFastaFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "afa")
    filenames = ["afa.good.1", "afa.good.2", "afa.good.3"]
    starts    = ["VLSEGEWQLV", "UCCGAUAUAG", "mqifvktltg"]
    format    = "afa"
    lengths   = [165, 73, 128]
    counts    = [4, 5, 6]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.rna(), easel.Alphabet.amino()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestClustalFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "clustal")
    filenames = ["clustal.good.2"]
    starts    = ["UCCGAUAUAG"]
    format    = "clustallike"
    lengths   = [73]
    counts    = [5]
    alphabet  = [easel.Alphabet.rna()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestClustalLikeFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "clustal")
    filenames = ["clustal.good.1"]
    starts    = ["VLSEGEWQLV"]
    format    = "clustallike"
    lengths   = [165]
    counts    = [4]
    alphabet  = [easel.Alphabet.amino()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestPhylipFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "phylip")
    filenames = ["phylip.good.1", "phylip.good.2", "phylip.good.3"]
    starts    = ["AAGCTNGGGC", "ATGGCGAAGG", "MKVILLFVLA"]
    format    = "phylip"
    lengths   = [42, 50, 384]
    counts    = [5, 7, 3]
    alphabet  = [easel.Alphabet.dna(), easel.Alphabet.dna(), easel.Alphabet.amino()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestPhylipsFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "phylips")
    filenames = ["phylips.good.1", "phylips.good.2"]
    starts    = ["AAGCTNGGGC", "MKVILLFVLA"]
    format    = "phylips"
    lengths   = [42, 384]
    counts    = [5, 3]
    alphabet  = [easel.Alphabet.dna(), easel.Alphabet.amino()]


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestPsiblastFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "psiblast")
    filenames = ["psiblast.good.1", "psiblast.good.2"]
    starts    = ["VLSEGEWQLV", "UCCGAUAUAG"]
    format    = "psiblast"
    lengths   = [165, 73]
    counts    = [4, 5]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.rna()]

    @unittest.expectedFailure
    def test_read_fileobject_guess_format(self):
        super().test_read_fileobject_guess_format()


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestSelexFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "selex")
    filenames = ["selex.good.1", "selex.good.2", "selex.good.3", "selex.good.4"]
    starts    = ["ACDEFGHIKL", "ACDEFGHIKL", "ACDEFGHIKL", "gGAGUAAGAU"]
    format    = "selex"
    lengths   = [20, 40, 52, 31]
    counts    = [5, 5, 7, 11]
    alphabet  = [easel.Alphabet.amino(), easel.Alphabet.amino(), easel.Alphabet.amino(), easel.Alphabet.rna()]

    @unittest.expectedFailure
    def test_read_fileobject_guess_format(self):
        super().test_read_fileobject_guess_format()


@unittest.skipUnless(os.path.exists(EASEL_FOLDER), "test data not available")
class TestStockholmFile(_TestReadFilename, _TestReadFileObject, unittest.TestCase):
    folder    = os.path.join(EASEL_FOLDER, "esl_msa_testfiles", "stockholm")
    filenames = ["stockholm.good.1"]
    starts    = ["ACDEFGHKLM"]
    format    = "stockholm"
    lengths   = [38]
    counts    = [7]
    alphabet  = [easel.Alphabet.amino()]



import pickle
import unittest
import sys

from pyhmmer import easel


class TestBitfield(unittest.TestCase):

    @unittest.skipIf(sys.implementation.name == "pypy", "`getsizeof` not supported on PyPY")
    def test_sizeof(self):
        bitfield = easel.Bitfield.zeros(8)
        self.assertGreater(sys.getsizeof(bitfield), 0)

    def test_buffer_protocol(self):
        bitfield = easel.Bitfield.zeros(8)
        self.assertEqual(memoryview(bitfield).tobytes()[0], 0b00000000)
        bitfield[0] = bitfield[2] = bitfield[3] = True
        self.assertEqual(memoryview(bitfield).tobytes()[0], 0b00001101)
        bitfield2 = easel.Bitfield.zeros(256)
        self.assertEqual(memoryview(bitfield2).shape[0], 4)

    def test_index_error(self):
        bitfield = easel.Bitfield.zeros(8)
        with self.assertRaises(IndexError):
            bitfield[9] = 1
        with self.assertRaises(IndexError):
            bitfield[-9] = 1

    def test_eq(self):
        b1 = easel.Bitfield.zeros(8)
        self.assertNotEqual(b1, object())

        b2 = easel.Bitfield.zeros(8)
        self.assertEqual(b1, b2)

        b3 = easel.Bitfield.zeros(7)
        self.assertNotEqual(b1, b3)

        b1.toggle(0)
        self.assertNotEqual(b1, b2)
        b2.toggle(0)
        self.assertEqual(b1, b2)

    def test_pickle(self):
        b1 = easel.Bitfield.zeros(8)
        b1[2] = b1[4] = True

        b2 = pickle.loads(pickle.dumps(b1))
        self.assertEqual(b1, b2)

        b3 = easel.Bitfield.zeros(8)
        self.assertNotEqual(b2, b3)

    def test_getstate(self):
        b1 = easel.Bitfield.zeros(8)
        b1[2] = b1[4] = True

        b2 = easel.Bitfield.zeros(8)
        b2[2] = b2[4] = True

        self.assertEqual(b1.__getstate__(), b2.__getstate__())

    def test_copy(self):
        b1 = easel.Bitfield.zeros(8)
        b1[2] = b1[4] = True

        b2 = b1.copy()
        for i in range(len(b1)):
            self.assertEqual(b1[i], b2[i])

        b2[2] = b2[4] = False
        self.assertTrue(b1[2])
        self.assertTrue(b1[4])
        self.assertFalse(b2[2])
        self.assertFalse(b2[4])


cdef extern from "esl_exponential.h" nogil:

    double esl_exp_pdf    (double x, double mu, double lambda);
    double esl_exp_logpdf (double x, double mu, double lambda);
    double esl_exp_cdf    (double x, double mu, double lambda);
    double esl_exp_logcdf (double x, double mu, double lambda);
    double esl_exp_surv   (double x, double mu, double lambda);
    double esl_exp_logsurv(double x, double mu, double lambda);
    double esl_exp_invcdf (double p, double mu, double lambda);
    double esl_exp_invsurv(double p, double mu, double lambda);


    double esl_exp_generic_pdf   (double x, void *params);
    double esl_exp_generic_cdf   (double x, void *params);
    double esl_exp_generic_surv  (double x, void *params);
    double esl_exp_generic_invcdf(double p, void *params);

    int    esl_exp_Plot(FILE *fp, double mu, double lambda,
    			   double (*func)(double x, double mu, double lambda),
    			   double xmin, double xmax, double xstep);

    double esl_exp_Sample(ESL_RANDOMNESS *r, double mu, double lambda);

    int esl_exp_FitComplete     (double *x, int n, double *ret_mu, double *ret_lambda);
    int esl_exp_FitCompleteScale(double *x, int n, double      mu, double *ret_lambda);

    int esl_exp_FitCompleteBinned(ESL_HISTOGRAM *h, double *ret_mu, double *ret_lambda);



cdef extern from "esl_composition.h" nogil:

    int esl_composition_BL62(double *f);
    int esl_composition_WAG (double *f);
    int esl_composition_SW34(double *f);
    int esl_composition_SW50(double *f);



# coding: utf-8
# cython: language_level=3
"""Implementation of a client for the HMMER daemon.

The HMMER daemon is a server daemon provided by HMMER3 to run distributed
search/scan pipelines on one or more worker machines. It is used to
power the `HMMER web server <https://www.ebi.ac.uk/Tools/hmmer/>`_.

This module only provides a client (such as the one in the ``hmmc2``
executable). To run the server, you need a database in the right format
(either a pressed HMM database, or sequences in a special FASTA format).
First launch the master process, and then the worker processes:

.. code:: console

    $ hmmpgmd --master --seqdb db.fasta --cport 51371 --wport 51372
    $ hmmpgmd --worker 127.0.0.1 --wport 51372

You can communicate with the master process after instantiating the
`pyhmmer.daemon.Client` class with the address of the master process and
the port on which it listens for client/server communication::

    >>> client = pyhmmer.daemon.Client("127.0.0.1", 51371)

"""

# --- C imports --------------------------------------------------------------

from cpython.bytearray cimport PyByteArray_AS_STRING

cimport libeasel
cimport libhmmer.hmmpgmd
cimport libhmmer.p7_hit
from libc.stdlib cimport free, realloc
from libc.string cimport memset, memcpy
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t
from libhmmer.hmmpgmd cimport HMMD_SEARCH_STATUS_SERIAL_SIZE, HMMD_SEARCH_STATUS, HMMD_SEARCH_STATS
from libhmmer.p7_pipeline cimport p7_pipemodes_e, P7_PIPELINE
from libhmmer.p7_hit cimport P7_HIT

cimport pyhmmer.plan7
from pyhmmer.easel cimport Sequence, Alphabet, MSA, KeyHash
from pyhmmer.errors import UnexpectedError, AllocationError, ServerError
from pyhmmer.plan7 cimport Builder, TopHits, Pipeline, HMM, Background

include "_getid.pxi"


# --- Python imports ---------------------------------------------------------

import io
import socket
import warnings


# --- Cython classes ---------------------------------------------------------

cdef str      DEFAULT_ADDRESS = "127.0.0.1"
cdef uint16_t DEFAULT_PORT    = 51371

cdef class Client:
    """A `socket`-based client to communicate with a HMMER daemon server.

    This class implements the client-side protocol to query a database with
    a `~pyhmmer.easel.Sequence`, `~pyhmmer.easel.MSA` or `~pyhmmer.plan7.HMM`.
    It must first connect to the server with the `~Client.connect` method::

        >>> client = daemon.Client("127.0.0.1", 51371)
        >>> client.connect()

    Afterwards, the client can be used to run pipelined searches,
    returning a `~pyhmmer.plan7.TopHits`::

        >>> client.search_hmm(thioesterase)
        <pyhmmer.plan7.TopHits object at 0x...>

    Additional keyword arguments can be passed to customize the pipelined
    search. All parameters from `~pyhmmer.plan7.Pipeline` are supported::

        >>> client.search_hmm(thioesterase, F1=0.02, E=1e-5)
        <pyhmmer.plan7.TopHits object at 0x...>

    Hint:
        `Client` implements the context manager protocol, which can be used
        to open and close a connection to the server within a context::

            >>> with daemon.Client() as client:
            ...    client.search_hmm(thioesterase)
            <pyhmmer.plan7.TopHits object at 0x...>

    Caution:
        Hits returned by the server will not have corresponding hit names,
        but only numerical identifiers. It is up to the client user to map
        these to the actual target names, often using an external file or
        database. If the database is small and several queries are made, it
        is feasible to parse the database from the client side to extract
        identifiers of the target HMMs or sequences.

    .. versionadded:: 0.6.0

    """

    # --- Magic methods ------------------------------------------------------

    def __init__(
        self,
        str address=DEFAULT_ADDRESS,
        uint16_t port=DEFAULT_PORT,
    ):
        """__init__(self, address="127.0.0.1", port=51371)\n--\n

        Create a new `Client` connecting to the given HMMER daemon server.

        Arguments:
            address (`str`): The address of the HMMER daemon server.
            port (`int`): The port over which the HMMER daemon server
                performs client/server communication.

        """
        self.address = address
        self.port = port
        self.socket = socket.socket()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_value, exc_type, traceback):
        self.close()

    def __repr__(self):
        cdef object ty   = type(self)
        cdef list   args = []
        if self.address != DEFAULT_ADDRESS:
            args.append(self.address)
        if self.port != DEFAULT_PORT:
            args.append(self.port)
        return "{}.{}({})".format(
            ty.__module__,
            ty.__name__,
            "".join(args)
        )

    # --- C Methods ----------------------------------------------------------

    cdef bytearray _recvall(self, size_t message_size):
        """Receive exactly ``message_size`` bytes from ``self.socket``.
        """
        cdef bytearray buffer = bytearray(message_size)
        cdef object    view   = memoryview(buffer)
        cdef size_t received  = 0
        cdef size_t recv_size = 0
        while received < message_size:
            recv_size = self.socket.recv_into(view)
            if recv_size == 0:
                raise EOFError(f"Expected message of size {message_size}, received {received}")
            received += recv_size
            view = view[recv_size:]
        return buffer

    cdef TopHits _client(
        self,
        object query,
        uint64_t db,
        list ranges,
        Pipeline pli,
        p7_pipemodes_e mode,
    ):
        """A generic implementation of the steps to communicate with the server.

        Arguments:
            query (`bytes`): A buffer storing the serialized query in
                encoded form.
            db (`int`): The index of the database to query.
            ranges (`list` of `tuple`): A list of ranges of target sequences
                to query inside the database.
            pli (`~plan7.Pipeline`): A pipeline object used to store generic
                configuration for the search/scan.
            mode (`int`): The pipeline mode marking whether the server should
                be queried in search or scan mode.

        """
        cdef int                status
        cdef HMMD_SEARCH_STATS  search_stats
        cdef HMMD_SEARCH_STATUS search_status

        cdef object             send_buffer
        cdef bytearray          response
        cdef const char*        response_data

        cdef uint32_t           hits_start
        cdef uint32_t           buf_offset    = 0
        cdef TopHits            hits          = TopHits(query)
        cdef str                options       = "".join(pli.arguments())

        # check ranges argument
        if ranges is not None:
            if len(ranges) < 1:
                raise ValueError("At least one range is needed for the `ranges` argument")
            elif any(len(r) != 2 for r in ranges):
                raise ValueError("`ranges` must be a list of two-element tuples")
            elif not all(isinstance(r[0], int) and isinstance(r[1], int) for r in ranges):
                raise TypeError("`ranges` must be a list where elements are 2-tuples of int")

        # clean memory for data structures allocated on the stack
        memset(&search_stats, 0, sizeof(HMMD_SEARCH_STATS))
        memset(&search_status, 0, sizeof(HMMD_SEARCH_STATUS))
        search_stats.hit_offsets = NULL

        # serialize query
        with io.BytesIO() as buffer:
            query.write(buffer)
            buffer.write(b"\n//")
            txt = buffer.getvalue()

        try:
            # send the options
            if mode == p7_pipemodes_e.p7_SEARCH_SEQS:
                if ranges is not None:
                    options = "--seqdb_ranges {} {}".format(
                        ",".join("{}..{}".format(*r) for r in ranges),
                        options
                    )
                self.socket.sendall(f"@--seqdb {db} {options}\n".encode("ascii"))
            else:
                self.socket.sendall(f"@--hmmdb {db} {options}\n".encode("ascii"))

            # send the query
            self.socket.sendall(txt)

            # get the search status back
            response = self._recvall(HMMD_SEARCH_STATUS_SERIAL_SIZE)
            status = libhmmer.hmmpgmd.hmmd_search_status_Deserialize(
                <const uint8_t*> PyByteArray_AS_STRING(response),
                &buf_offset,
                &search_status
            )
            if status != libeasel.eslOK:
                raise UnexpectedError(status, "hmmd_search_status_Deserialize")

            # check if error happened
            if search_status.status != libeasel.eslOK:
                error = self.socket.recv(search_status.msg_size)
                raise ServerError(search_status.status, error.decode("utf-8", "replace"))

            # get the response
            response = self._recvall(search_status.msg_size)
            response_data = PyByteArray_AS_STRING(response)

            with nogil:
                # deserialize search_stats
                buf_offset = 0
                status = libhmmer.hmmpgmd.p7_hmmd_search_stats_Deserialize(
                    <const uint8_t*> response_data,
                    &buf_offset,
                    &search_stats
                )
                if status != libeasel.eslOK:
                    raise UnexpectedError(status, "p7_hmmd_search_search_stats_Deserialize")

                # copy input parameters from the pipeline
                memcpy(&hits._pli, pli._pli, sizeof(P7_PIPELINE))

                # copy the search search_stats
                hits._pli.mode                = mode
                hits._pli.nmodels             = search_stats.nmodels
                hits._pli.nseqs               = search_stats.nseqs
                hits._pli.n_past_msv          = search_stats.n_past_msv
                hits._pli.n_past_vit          = search_stats.n_past_vit
                hits._pli.n_past_fwd          = search_stats.n_past_fwd
                hits._pli.Z                   = search_stats.Z
                hits._pli.domZ                = search_stats.domZ
                hits._pli.Z_setby             = search_stats.Z_setby
                hits._pli.domZ_setby          = search_stats.domZ_setby
                hits._th.nreported            = search_stats.nreported
                hits._th.nincluded            = search_stats.nincluded
                hits._th.is_sorted_by_seqidx  = False
                hits._th.is_sorted_by_sortkey = True
                hits._th.N                    = search_stats.nhits

                # reallocate hit arrays
                if search_stats.nhits > 0:
                    hits._th.unsrt = <P7_HIT*> realloc(hits._th.unsrt, search_stats.nhits * sizeof(P7_HIT))
                    if hits._th.unsrt == NULL:
                        raise AllocationError("P7_HIT", sizeof(P7_HIT), search_stats.nhits)
                    hits._th.hit = <P7_HIT**> realloc(hits._th.hit, search_stats.nhits * sizeof(P7_HIT*))
                    if hits._th.hit == NULL:
                        raise AllocationError("P7_HIT*", sizeof(P7_HIT*), search_stats.nhits)

                # deserialize hits
                hits_start = buf_offset
                for i in range(search_stats.nhits):
                    # clean pointers in hit data to force reallocation
                    hits._th.unsrt[i].name = NULL
                    hits._th.unsrt[i].acc  = NULL
                    hits._th.unsrt[i].desc = NULL
                    hits._th.unsrt[i].dcl  = NULL
                    # check buffers match the ones in the search stats
                    if buf_offset - hits_start != search_stats.hit_offsets[i]:
                        with gil:
                            warnings.warn(f"Hit offset {i} did not match expected (expected {search_stats.hit_offsets[i]}, found {buf_offset - hits_start})")
                    # deserialize and record the hit
                    status = libhmmer.p7_hit.p7_hit_Deserialize(
                        <const uint8_t*> response_data,
                        &buf_offset,
                        &hits._th.unsrt[i]
                    )
                    if status != libeasel.eslOK:
                        raise UnexpectedError(status, "p7_hit_Deserialize")
                    hits._th.hit[i] = &hits._th.unsrt[i]

        finally:
            free(search_stats.hit_offsets)

        return hits

    # --- Python Methods -----------------------------------------------------

    def connect(self):
        """Connect the client to the HMMER daemon server.
        """
        self.socket.connect((self.address, self.port))

    def close(self):
        """Close the connection to the HMMER daemon server.
        """
        self.socket.close()

    def search_seq(
        self,
        Sequence query,
        uint64_t db = 1,
        list ranges = None,
        **options
    ):
        """Search the HMMER daemon database with a query sequence.

        Arguments:
            query (`~pyhmmer.easel.Sequence`): The sequence object to use
                to query the sequence database.
            db (`int`): The index of the sequence database to query.
            ranges (`list` of `tuple`): A list of ranges of target sequences
                to query inside the database.

        Returns:
            `~pyhmmer.plan7.TopHits`: The hits found in the target sequences.

        Hint:
            This method corresponds to running ``phmmer`` with the ``query``
            sequence against the sequence database loaded on the server side.

        """
        cdef Alphabet abc  = getattr(query, "alphabet", Alphabet.amino())
        cdef Pipeline pli  = Pipeline(abc, **options)
        return self._client(query, db, ranges, pli, p7_pipemodes_e.p7_SEARCH_SEQS)

    def search_hmm(
        self,
        HMM query,
        uint64_t db = 1,
        list ranges = None,
        **options
    ):
        """Search the HMMER daemon database with a query HMM.

        Arguments:
            query (`~pyhmmer.easel.MSA`): The profile HMM object to use to
                query the sequence database.
            db (`int`): The index of the sequence database to query.
            ranges (`list` of `tuple`): A list of ranges of target sequences
                to query inside the database.

        Returns:
            `~pyhmmer.plan7.TopHits`: The hits found in the target sequences.

        Hint:
            This method corresponds to running ``hmmsearch`` with the
            ``query`` HMM against the sequence database loaded on the
            server side.

        """
        cdef Pipeline pli  = Pipeline(query.alphabet, **options)
        return self._client(query, db, ranges, pli, p7_pipemodes_e.p7_SEARCH_SEQS)

    def scan_seq(self, Sequence query, uint64_t db = 1, **options):
        """Search the HMMER daemon database with a query sequence.

        Arguments:
            query (`~pyhmmer.easel.Sequence`): The sequence object to use
                to query the HMM database.
            db (`int`): The index of the HMM database to query.

        Returns:
            `~pyhmmer.plan7.TopHits`: The hits found in the target sequences.

        Hint:
            This method corresponds to running ``hmmscan`` with the
            ``query`` sequence against the HMM database loaded on the
            server side.

        """
        cdef Alphabet abc  = getattr(query, "alphabet", Alphabet.amino())
        cdef Pipeline pli  = Pipeline(abc, **options)
        return self._client(query, db, None, pli, p7_pipemodes_e.p7_SCAN_MODELS)

    def iterate_seq(
        self,
        Sequence query,
        uint64_t db = 1,
        list ranges = None,
        Builder builder = None,
        object select_hits = None,
        **options
    ):
        """Search iteratively against the daemon database with a query sequence.

        Arguments:
            query (`~pyhmmer.easel.Sequence`): The sequence object to use
                to query the server-side sequence database.
            db (`int`): The index of the sequence database to query.
            ranges (`list` of `tuple`, optional): A list of ranges of target
                sequences to query inside the database.
            builder (`~pyhmmer.plan7.Builder`, optional): A builder instance
                to use to convert the `~pyhmmer.easel.MSA` objects obtained
                during each iteration into `~pyhmmer.plan7.HMM` objects.
            select_hits (callable, optional): A function or callable object
                for manually selecting hits during each iteration. It should
                take a single `TopHits` argument and change the inclusion of
                individual hits with the `Hit.include` and `Hit.drop` methods.

        Returns:
            `~pyhmmer.daemon.IterativeSearch`: An iterator object yielding
            the hits, sequence alignment, and HMM for each iteration.

        Hint:
            This method corresponds to running ``jackhmmer`` with the
            ``query`` sequence against the sequence database loaded on the
            server side.

        Caution:
            Default values used for ``jackhmmer`` do not correspond to the
            default parameters used for creating a pipeline in the other
            cases. This method will use default values of ``incE=0.001`` and
            ``incdomE=0.001`` unless other values are given as keyword
            arguments.

        """
        # check that builder is in hand architecture, not fast
        if builder is None:
            builder = Builder(Alphabet.amino(), architecture="hand")
        elif builder.architecture != "hand":
            raise ValueError("`iterate_seq` only supports a builder with 'hand' architecture")
        # use `jackhmmer` defaults
        options.setdefault("incE", 0.001)
        options.setdefault("incdomE", 0.001)
        # return the iterator
        return IterativeSearch(self, query, db, builder, ranges, select_hits, options)

    def iterate_hmm(
        self,
        HMM query,
        uint64_t db = 1,
        list ranges = None,
        Builder builder = None,
        object select_hits = None,
        **options
    ):
        """Search iteratively against the daemon database with a query HMM.

        Arguments:
            query (`~pyhmmer.plan7.HMM`): The HMM object to use to query the
                server-side sequence database.
            db (`int`): The index of the sequence database to query.
            ranges (`list` of `tuple`, optional): A list of ranges of target
                sequences to query inside the database.
            builder (`~pyhmmer.plan7.Builder`, optional): A builder instance
                to use to convert the `~pyhmmer.easel.MSA` objects obtained
                during each iteration into `~pyhmmer.plan7.HMM` objects.
            select_hits (callable, optional): A function or callable object
                for manually selecting hits during each iteration. It should
                take a single `TopHits` argument and change the inclusion of
                individual hits with the `Hit.include` and `Hit.drop` methods.

        Returns:
            `~pyhmmer.daemon.IterativeSearch`: An iterator object yielding
            the hits, sequence alignment, and HMM for each iteration.

        Hint:
            This method corresponds to running ``jackhmmer`` with the
            ``query`` HMM against the sequence database loaded on the
            server side.

        Caution:
            Default values used for ``jackhmmer`` do not correspond to the
            default parameters used for creating a pipeline in the other
            cases. This method will use default values of ``incE=0.001`` and
            ``incdomE=0.001`` unless other values are given as keyword
            arguments.

        """
        # check that builder is in hand architecture, not fast
        if builder is None:
            builder = Builder(Alphabet.amino(), architecture="hand")
        elif builder.architecture != "hand":
            raise ValueError("`iterate_seq` only supports a builder with 'hand' architecture")
        # use `jackhmmer` defaults
        options.setdefault("incE", 0.001)
        options.setdefault("incdomE", 0.001)
        # return the iterator
        return IterativeSearch(self, query, db, builder, ranges, select_hits, options)


cdef class IterativeSearch(pyhmmer.plan7.IterativeSearch):
    """A helper class for running iterative queries against a HMMER daemon.

    See `Client.iterate_seq` and `Client.iterate_hmm` for more
    information.

    Attributes:
        client (`~pyhmmer.daemon.Client`): The HMMER daemon client to use
            to get hits on each iteration.
        db (`int`): The index of the database to query for hits on each
            iteration.
        builder (`~pyhmmer.plan7.Builder`): The builder object for
            converting multiple sequence alignments obtained after each
            run to a `~pyhmmer.plan7.HMM`.
        query (`~pyhmmer.easel.DigitalSequence` or `~pyhmmer.plan7.HMM`):
            The query object to use for the first iteration.
        converged (`bool`): Whether the iterative search already converged
            or not.
        ranking (`~pyhmmer.easel.KeyHash`): A mapping storing the rank of
            hits from previous iterations.
        iteration (`int`): The index of the last iteration done so far.

    Yields:
        `~pyhmmer.plan7.IterationResult`: A named tuple containing the hits,
        multiple sequence alignment and HMM for each iteration, as well as
        the iteration index and a flag marking whether the search converged.

    References:
        - Johnson, Steven L., Eddy, Sean R. & Portugaly, Elon.
          *Hidden Markov model speed heuristic and iterative HMM search
          procedure*. BMC Bioinformatics 11, 431 (18 August 2010).
          :doi:`10.1186/1471-2105-11-431`.

    .. versionadded:: 0.6.0

    """

    def __init__(
        self,
        Client client,
        object query,
        uint64_t db,
        Builder builder,
        list ranges = None,
        object select_hits = None,
        dict options = None
    ):
        """__init__(self, client, query, db, builder, ranges=None, select_hits=None, options=None)\n--\n

        Create a new iterative search against the HMMER daemon.

        Hint:
            Use the `Client.iterate_hmm` method instead of creating this
            object directly.

        """
        self.client = client
        self.builder = builder
        self.select_hits = select_hits
        self.options = options or {}
        self.db = db
        self.query = query
        self.background = Background(builder.alphabet)
        self.converged = False
        self.ranking = KeyHash()
        self.msa = None
        self.iteration = 0
        self.pipeline = None
        self.targets = None

    cpdef TopHits _search_hmm(self, HMM hmm):
        return self.client.search_hmm(
            hmm,
            db=self.db,
            ranges=self.ranges,
            **self.options
        )



cdef inline size_t new_capacity(size_t capacity, size_t length) nogil:
    """new_capacity(capacity, length)\n--

    Compute a new capacity for a buffer reallocation.

    """
    return (<size_t> capacity + (capacity >> 3) + 6) & ~(<size_t> 3)
 



from libc.stdio cimport FILE

from libeasel cimport ESL_DSQ
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.hmm cimport ESL_HMM


cdef extern from "hmmer.h" nogil:

    ctypedef p7_bg_s P7_BG
    cdef struct p7_bg_s:
        float* f
        float p1
        ESL_HMM* fhmm
        float omega
        const ESL_ALPHABET* abc

    P7_BG *p7_bg_Create(const ESL_ALPHABET *abc)
    P7_BG *p7_bg_CreateUniform(const ESL_ALPHABET *abc)
    P7_BG *p7_bg_Clone(const P7_BG *bg)
    int    p7_bg_Dump(FILE *ofp, const P7_BG *bg)
    void   p7_bg_Destroy(P7_BG *bg)
    int    p7_bg_SetLength(P7_BG *bg, int L)
    int    p7_bg_NullOne(const P7_BG *bg, const ESL_DSQ *dsq, int L, float *ret_sc)

    int    p7_bg_Read(char *bgfile, P7_BG *bg, char *errbuf)
    int    p7_bg_Write(FILE *fp, P7_BG *bg)

    int    p7_bg_SetFilter  (P7_BG *bg, int M, const float *compo)
    int    p7_bg_FilterScore(P7_BG *bg, const ESL_DSQ *dsq, int L, float *ret_sc)
